# FILE: any2ppm.tcl
#
# This script converts data in any image format understood by Tk's
# (possibly extended) [image photo] command to data representing the
# same image in the PPM P3 format.
#
# This file must be interpreted by the filtersh shell.

package require Oc 1.1
package require Tk
Oc_ForceStderrDefaultMessage

Oc_Main SetAppName any2ppm
Oc_Main SetVersion 1.2.0.3

Oc_CommandLine Option console {} {}

Oc_CommandLine Option noinfo {} {
    global noinfo;  set noinfo 1
} {Suppress writing of progress info to stderr}
set noinfo 0

Oc_CommandLine Option o {outfile} {
    global outSpec; set outSpec $outfile
} {Write output to outfile ("-" for stdout)}
set outSpec ""

Oc_CommandLine Option [Oc_CommandLine Switch] {
    {{infile list} {} {Input file(s).  If none or "", read from stdin.}}
    } {
    global inList; set inList $infile
} {End of options; next argument is infile}
set inList [list]

Oc_CommandLine Parse $argv

set errcount 0
set BUFSIZ 250000       ;# Size of output write buffer

proc make_ppm_name { inname } {
    set outname [file rootname $inname].ppm
    if {[file exists $outname]} {
	set basename $outname
	set outname {}
	for {set i 1} {$i<100} {incr i} {
	    set testname ${basename}-$i
	    if {![file exists $testname]} {
		set outname $testname
		break
	    }
	}
    }
    return $outname
}

proc write_ppm_file { chanid pic } {
    # The Tk photo interface does not have a "write to channel" option.
    set width [image width $pic] ; set height [image height $pic]
    puts $chanid "P3"
    puts $chanid "$width $height"
    puts $chanid "255"
    set rtop [expr {$height-1}] ; set ctop [expr {$width-1}]
    for {set r 0} {$r<$height} {incr r} {
	for {set c 0} {$c<$width} {incr c} {
	    puts $chanid [$pic get $c $r]
	}
    }
    flush $chanid
}

switch -exact -- $outSpec {
    {} {set outmode automatic}
    {-} {
	set outmode fixed
	set outchan stdout
	fconfigure $outchan -translation auto \
		-buffering full -buffersize $BUFSIZ
    }
    default {
	set outmode fixed
	set outname $outSpec
	if {[catch {open $outname "w"} msg]} {
	    puts stderr "Unable to open output file $outname: $msg"
	    incr errcount
	    exit $errcount
	} else {
	    set outchan $msg
	    fconfigure $outchan -translation auto \
		    -buffering full -buffersize $BUFSIZ
	}
    }
}

set loopOpen 0
if {[llength $inList]==0} {
    set inList [list {}]  ;# Use empty string to denote stdin
}
foreach inname $inList {
    if {[string match {} $inname]} {
	# Read from stdin
	if {!$noinfo} { puts stderr "Processing input from stdin" }
	fconfigure stdin -translation binary
	set data [read stdin]
	if {[catch {set pic [image create photo -data $data]} errmsg]} {
	    puts stderr "ERROR: $errmsg"
	    incr errcount
	    continue
	}
    } else {
	# Read from file
	if {![file readable $inname]} {
	    puts stderr "Unable to open input file $inname; Skipping."
	    incr errcount
	    continue
	}
	if {!$noinfo} { puts stderr "Processing input file $inname" }
	if {[catch {set pic [image create photo -file $inname]} errmsg]} {
	    puts stderr "ERROR: $errmsg"
	    incr errcount
	    continue
	}
    }

    if {!$noinfo} { puts stderr "Processing output..." }
    if {[string match automatic $outmode]} {
	if {[string match {} $inname]} {
	    set outchan stdout
	} else {
	    set outname [make_ppm_name $inname]
	    if {[string match {} $outname]} {
		puts stderr "Unable to generate unique output name; Skipping."
		incr errcount
		continue
	    }
	    if {[catch {open $outname "w"} outchan]} {
		puts stderr "Unable to open output file $outname; Skipping."
		incr errcount
		continue	    
	    }
	    fconfigure $outchan -translation auto \
		    -buffering full -buffersize $BUFSIZ
	    set loopOpen 1
	}
    }

    if {[catch {write_ppm_file $outchan $pic} errmsg]} {    
	puts stderr "PPM WRITE ERROR: $errmsg"
	incr errcount
    }
    flush $outchan
    if {!$noinfo} {
	if {[info exists outname]} {
	    puts stderr "Output written to $outname"
	} elseif {[string match stdout $outchan]} {
	    puts stderr "Output written to stdout"
	} else {
	    puts stderr "Unknown output (programming error?)"
	}
    }

    if {$loopOpen} {
	close $outchan
	unset outchan
    }
}

exit $errcount
