/* FILE: driver.cc                 -*-Mode: c++-*-
 *
 * Abstract driver class; fills states and initiates steps.
 * Also registers the command "Oxs_Run" with the Tcl intepreter.
 *
 */

#include "oc.h"
#include "nb.h"
#include "director.h"
#include "driver.h"
#include "simstate.h"
#include "key.h"
#include "energy.h"	// Needed to make MSVC++ 5 happy

#include <string>
#include <vector>

/* End includes */

// Constructor
Oxs_Driver::Oxs_Driver
( const char* name,        // Child instance id
  Oxs_Director* newdtr,    // App director
  const char* argstr       // Args
  ) : Oxs_Ext(name,newdtr,argstr),
  attributes("Oxs_LabelValue",newdtr,""),
  number_of_stages(1)
{
  // Reserve state space in director
  director->ReserveSimulationStateRequest(2);

  // Set attribute strings
  string basename = GetStringInitValue("basename","oxs");
  string fieldformat = GetStringInitValue("vector_field_output_format",
					  "binary 8");
  string scalarformat = GetStringInitValue("scalar_output_format",
					   "%.17g");
  attributes.SetLabel("basename",basename);
  attributes.SetLabel("vector_field_output_format",fieldformat);
  attributes.SetLabel("scalar_output_format",scalarformat);

  // Iteration count and stage limits
  if(HasInitValue("stage_iteration_limit")) {
    // Optional; default is no control
    GetGroupedUIntListInitValue("stage_iteration_limit",
				stage_iteration_limit);
  } else {
    stage_iteration_limit.push_back(0); // Default: no control
  }
  total_iteration_limit = GetUIntInitValue("total_iteration_limit",0);
  stage_count_request = GetUIntInitValue("stage_count",0);
  stage_count_check = GetIntInitValue("stage_count_check",1);
}

//Destructor
Oxs_Driver::~Oxs_Driver()
{}

// Oxs_Driver version of Init().  All children of Oxs_Driver *must*
// call this function in their Init() routines.  The main purpose
// of this function is to setup base driver outputs and to initialize
// the current state.
#define OSO_INIT(name,descript,units) \
   name##_output.Setup(this,InstanceName(),descript,units,0, \
           &Oxs_Driver::Fill__##name##_output,director)
BOOL Oxs_Driver::Init()
{
  BOOL success=1;
  OSO_INIT(iteration_count,"Iteration","");
  OSO_INIT(stage_iteration_count,"Stage iteration","");
  OSO_INIT(stage_number,"Stage","");
  spin_output.Setup(this,InstanceName(),"Spin","",1,
		    &Oxs_Driver::Fill__spin_output,
		    director);
  magnetization_output.Setup(this,InstanceName(),
			     "Magnetization","A/m",1,
			     &Oxs_Driver::Fill__magnetization_output,
			     director);
  aveMx_output.Setup(this,InstanceName(),"Mx","A/m",1,
		     &Oxs_Driver::Fill__aveM_output,
		     director);
  aveMy_output.Setup(this,InstanceName(),"My","A/m",1,
		     &Oxs_Driver::Fill__aveM_output,
		     director);
  aveMz_output.Setup(this,InstanceName(),"Mz","A/m",1,
		     &Oxs_Driver::Fill__aveM_output,
		     director);
  maxSpinAng_output.Setup(this,InstanceName(),"Max Spin Ang","deg",1,
		     &Oxs_Driver::Fill__maxSpinAng_output,
		     director);
  stage_maxSpinAng_output.Setup(this,InstanceName(),
		     "Stage Max Spin Ang","deg",1,
		     &Oxs_Driver::Fill__maxSpinAng_output,
		     director);
  run_maxSpinAng_output.Setup(this,InstanceName(),
                     "Run Max Spin Ang","deg",1,
		     &Oxs_Driver::Fill__maxSpinAng_output,
		     director);

  // Adjust spin output to always use full precision
  string default_format = spin_output.GetOutputFormat();
  Oxs_SplitList arglist;
  arglist.Split(default_format.c_str());
  if(arglist.Count()!=2) {
    OXS_THROW(Oxs_ProgramLogicError,
	      "Wrong number of arguments in spin output format string, "
	      "detected in Oxs_Driver Init");
  } else {
    vector<string> sarr;
    sarr.push_back(arglist[0]); // Data type
    if(sarr[0].compare("binary") == 0) {
      sarr.push_back("8");      // Precision
    } else {
      sarr.push_back("%.17g");
    }
    string precise_format = Oxs_MergeList(&sarr);
    spin_output.SetOutputFormat(precise_format.c_str());
  }

  // Determine total stage count requirements
  unsigned int min,max;
  director->ExtObjStageRequestCounts(min,max);
  if(stage_count_check!=0 && min>max) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Stage count request incompatibility detected;"
		" request range is [%u,%u].  Double check stage"
		" lists and applied field specifications.  The"
		" stage count compatibility check may be disabled"
		" in the driver Specify block by setting"
		" stage_count_check to 0.",
		min,max);
    throw Oxs_Ext::Error(this,buf);
  }

  // Parameter stage_count_request overrides all automatic settings
  // if set to a value different from 0.  Otherwise, use maximal
  // "min" value requested by all ext objects, unless that value is
  // zero, in which case we use a 1 stage default.
  if(stage_count_request>0) number_of_stages = stage_count_request;
  else                      number_of_stages = min;
  if(number_of_stages<1) number_of_stages=1; // Default.

  // Initialize current state from initial state provided by
  // concrete child class.
  current_state = GetInitialState();
  if (current_state.GetPtr() == NULL) {
    success = 0; // Error.  Perhaps an exception throw would be better?
  } else {
    current_state.GetReadReference();
  }

  return success;
}

void Oxs_Driver::StageRequestCount(unsigned int& min,
				   unsigned int& max) const
{
  min = stage_iteration_limit.size();
  if(min>1) max = min;
  else      max = UINT_MAX;

  if(stage_count_request>min) min = stage_count_request;
  if(stage_count_request>0 && stage_count_request<max) {
    max = stage_count_request;
  }
}

BOOL Oxs_Driver::IsStageDone(const Oxs_SimState& state) const
{
  if(state.stage_done == Oxs_SimState::DONE) return 1;
  if(state.stage_done == Oxs_SimState::NOT_DONE) return 0;
  /// Otherwise, state.stage_done == Oxs_SimState::UNKNOWN

  // Check state against parent Oxs_Driver class stage limiters.
  if(total_iteration_limit > 0
     && state.iteration_count >= total_iteration_limit ) {
    state.stage_done = Oxs_SimState::DONE;
    return 1;
  }

  // Stage iteration check
  UINT4m stop_iteration=0;
  if(state.stage_number >= stage_iteration_limit.size()) {
    stop_iteration
      = stage_iteration_limit[stage_iteration_limit.size()-1];
  } else {
    stop_iteration = stage_iteration_limit[state.stage_number];
  }
  if(stop_iteration>0
     && state.stage_iteration_count + 1 >= stop_iteration) {
    // Note: stage_iteration_count is 0 based, so the number
    // of iterations is stage_iteration_count + 1.
    state.stage_done = Oxs_SimState::DONE;
    return 1;
  }

  // Otherwise, leave it up to the child
  if(ChildIsStageDone(state)) {
    state.stage_done = Oxs_SimState::DONE;
    return 1;
  }

  state.stage_done = Oxs_SimState::NOT_DONE;
  return 0;
}

BOOL Oxs_Driver::IsRunDone(const Oxs_SimState& state) const
{
  if(state.run_done == Oxs_SimState::DONE) return 1;
  if(state.run_done == Oxs_SimState::NOT_DONE) return 0;
  /// Otherwise, state.run_done == Oxs_SimState::unknown

  // Check state against parent Oxs_Driver class run limiters.
  if(total_iteration_limit > 0
     && state.iteration_count >= total_iteration_limit) {
    state.run_done = Oxs_SimState::DONE;
    return 1;
  }

  if(number_of_stages > 0) {
    if( state.stage_number >= number_of_stages ||
	(state.stage_number+1 == number_of_stages
	&& IsStageDone(state))) {
      state.run_done = Oxs_SimState::DONE;
      return 1;
    }
  }	 

  // Otherwise, leave it up to the child
  if(ChildIsRunDone(state)) {
    state.run_done = Oxs_SimState::DONE;
    return 1;
  }

  state.run_done = Oxs_SimState::NOT_DONE;
  return 0;
}

void Oxs_Driver::FillState
(const Oxs_SimState& old_state,
 Oxs_SimState& new_state) const
{
  new_state.previous_state_id = old_state.Id();
  new_state.stage_number = old_state.stage_number;
  new_state.stage_start_time = old_state.stage_start_time;
  new_state.Ms           = old_state.Ms;
  new_state.Ms_inverse   = old_state.Ms_inverse;
  new_state.mesh         = old_state.mesh;
  new_state.spin.AdjustSize(new_state.mesh);
  new_state.stage_done = Oxs_SimState::UNKNOWN;
  new_state.run_done = Oxs_SimState::UNKNOWN;

  // State variables iteration_count, stage_iteration_count,
  // stage_elapsed_time, last_timestep and DerivedData should
  // be set elsewhere.  Initialize to dummy values to aid
  // fault detection.
  new_state.iteration_count
    = new_state.stage_iteration_count
    = static_cast<UINT4m>(-1);
  new_state.stage_elapsed_time
    = new_state.last_timestep
    = -1.0;
  new_state.ClearDerivedData();
  // Might want to add energy & old_energy to all states, to
  // simplify Delta_E rendering.  Or maybe every state has a
  // previous_state_energy parallel to previous_state_id, and
  // a Energy entry in DerivedData?  See also FillNewStageState.

  // Carry across old stage and run maxang data
  REAL8m run_maxang = -1.;
  if(!old_state.GetDerivedData("Run Max Spin Ang",run_maxang)) {
    // Spin angle data not filled.  Insure that maxang is
    // computed every step.  This is a little extra work; if we
    // wanted, we could coordinate this with the solver so that
    // if maxang at previous step + 2*(allowed spin movement)
    // was less than stage_maxang, then maxang would not need
    // to be computed this step.  This could be handled by
    // adding a maxang_guess value to the state, but it is
    // probably not worth the trouble.
    UpdateSpinAngleData(old_state);
    old_state.GetDerivedData("Run Max Spin Ang",run_maxang);
  }
  new_state.AddDerivedData("PrevState Run Max Spin Ang",run_maxang);
  REAL8m stage_maxang = -1.;
  if(!old_state.GetDerivedData("Stage Max Spin Ang",stage_maxang)) {
    UpdateSpinAngleData(old_state);
    old_state.GetDerivedData("Stage Max Spin Ang",stage_maxang);
  }
  new_state.AddDerivedData("PrevState Stage Max Spin Ang",stage_maxang);
}

void Oxs_Driver::FillNewStageState
(const Oxs_SimState& old_state,
 int new_stage_number,
 Oxs_SimState& new_state) const
{
  if(new_stage_number<0 ||
     (0<number_of_stages
      && number_of_stages<static_cast<UINT4m>(new_stage_number))) {
    // Valid stage numbers are 0, 1, ..., number_of_stages,
    // where the last indicates a "Run Done" state.
    char buf[1024];
    if(number_of_stages>0) {
      Oc_Snprintf(buf,sizeof(buf),
		  "Invalid stage request: %d;"
		  " should be in range [0,%d] (inclusive)",
		  new_stage_number,number_of_stages);
    } else {
      Oc_Snprintf(buf,sizeof(buf),
		  "Invalid stage request: %d; must be non-negative",
		  new_stage_number);
    }
    throw Oxs_Ext::Error(this,buf);
  }
  new_state = old_state; // Copy data
  new_state.previous_state_id = old_state.Id();
  new_state.iteration_count = old_state.iteration_count + 1;
  /// Increment iteration count across stage boundary.
  new_state.stage_number = new_stage_number;
  new_state.stage_iteration_count=0;
  new_state.stage_start_time += new_state.stage_elapsed_time;
  new_state.stage_elapsed_time = 0.0;
  new_state.stage_done = Oxs_SimState::UNKNOWN;
  new_state.run_done = Oxs_SimState::UNKNOWN;
  new_state.ClearDerivedData(); // Derived data no longer valid
  // Might want to add energy & old_energy to all states, to
  // simplify Delta_E rendering.  Or maybe every state has a
  // previous_state_energy parallel to previous_state_id, and
  // a Energy entry in DerivedData?  See also FillNewState.

  // Carry across run maxang data, and
  // initialize prevstate stage maxang to 0.
  REAL8m run_maxang = -1.0;
  if(!old_state.GetDerivedData("Run Max Spin Ang",run_maxang)) {
    // Spin angle data not filled in current state
    UpdateSpinAngleData(old_state); // Update
    old_state.GetDerivedData("Run Max Spin Ang",run_maxang);
  }
  new_state.AddDerivedData("PrevState Run Max Spin Ang",run_maxang);
  new_state.AddDerivedData("PrevState Stage Max Spin Ang",0.0);
}


void Oxs_Driver::SetStage(UINT4m requestedStage)
{
  if(current_state.GetPtr() == NULL) {
    // Current state is not initialized.
    string msg="Current state in Oxs_Driver is not initialized;"
      " This is probably the fault of the child class "
      + string(ClassName());
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  // Bound checks.  Should this return an error?  For now we'll
  // just adjust to bound.
  if(0<number_of_stages && requestedStage >= number_of_stages ) {
    requestedStage = number_of_stages - 1;
  }

  const Oxs_SimState& cstate = current_state.GetReadReference();
  if(requestedStage != cstate.stage_number) {
    // Change state only if stage number changes.
    Oxs_Key<Oxs_SimState> next_state;
    director->GetNewSimulationState(next_state);

    Oxs_SimState& nstate = next_state.GetWriteReference();
    FillNewStageState(cstate,requestedStage,nstate);
    next_state.GetReadReference();  // Release write lock

    current_state = next_state;
    current_state.GetReadReference(); // Hold read lock
  }
}

UINT4m Oxs_Driver::GetStage() const
{
  const Oxs_SimState* cstate = current_state.GetPtr();
  if(cstate == NULL) {
    // Current state is not initialized.
    string msg="Current state in Oxs_Driver is not initialized;"
      " This is probably the fault of the child class "
      + string(ClassName());
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  return cstate->stage_number;
}

void Oxs_Driver::Run(vector<OxsRunEvent>& results)
{ // Called by director
  results.clear();

  if(current_state.GetPtr() == NULL) {
    // Current state is not initialized.
    string msg="Current state in Oxs_Driver is not initialized;"
      " This is probably the fault of the child class "
      + string(ClassName());
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  if(current_state.ObjectId()==0) {
    // Current state is not fixed, i.e., is incomplete or transient.
    // To some extent, this check is not necessary, because key should
    // throw an exception on GetReadReference if the pointed to Oxs_Lock
    // object isn't fixed.
    string msg="PROGRAMMING ERROR:"
      " Current state in Oxs_Driver is incomplete or transient;"
      " This is probably the fault of the child class "
      + string(ClassName());
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  if (IsRunDone(current_state.GetReadReference())) {
    results.push_back(OxsRunEvent(OXS_RUN_DONE_EVENT,current_state));
    return;
  }

  // Try to take a step.  Currently, we return after one attempt,
  // but in the future we may use more complex criteria
  while (1) {
    if(IsStageDone(current_state.GetReadReference())) {
      // At end of current stage.  Move state across
      // stage boundary.
      Oxs_Key<Oxs_SimState> next_state;
      director->GetNewSimulationState(next_state);
      const Oxs_SimState& cstate = current_state.GetReadReference();
      Oxs_SimState& nstate = next_state.GetWriteReference();
      FillNewStageState(cstate,cstate.stage_number+1,nstate);
      next_state.GetReadReference();  // Release write lock
      current_state = next_state; // Frees old read lock
      results.push_back(OxsRunEvent(OXS_STEP_EVENT,current_state));
    } else {
      // Not at end of current stage.  Attempt step.
      Oxs_Key<Oxs_SimState> next_state;
      director->GetNewSimulationState(next_state);
      FillState(current_state.GetReadReference(),
		next_state.GetWriteReference());
      // NOTE: At this point next_state holds a write lock.
      //   The Step() function can make additional calls
      //   to next_state.GetWriteReference() as needed; write
      //   locks do not accumulate.  However, it is the
      //   resposibility of Step or its callees to release
      //   the write lock, once next_state is fully populated.
      if( Step(current_state,next_state) ) {
	// Good step.  Release read lock on old current_state,
	// and copy key from next_state.
	next_state.GetReadReference();  // Safety write lock release
	current_state = next_state; // Free old read lock
	UpdateSpinAngleData(*(current_state.GetPtr())); // Update
	/// max spin angle data on each accepted step.  Might want
	/// to modify this to instead estimate max angle change,
	/// and only do actually calculation when estimate uncertainty
	/// gets larger than some specified value.
	results.push_back(OxsRunEvent(OXS_STEP_EVENT,current_state));
      }
    }

    // Check if stage or problem done is true
    if (IsStageDone(current_state.GetReadReference())) {
      results.push_back(OxsRunEvent(OXS_STAGE_DONE_EVENT,current_state));
    }
    if (IsRunDone(current_state.GetReadReference())) {
      results.push_back(OxsRunEvent(OXS_RUN_DONE_EVENT,current_state));
    }

    // For now, always break out of loop.  We may change this
    // in the future.
    break;
  }

}

////////////////////////////////////////////////////////////////////////
// State-based outputs, maintained by the driver.  These are
// conceptually public, but are specified private to force
// clients to use the output_map interface in Oxs_Director.

#define OSO_FUNC(NAME) \
void Oxs_Driver::Fill__##NAME##_output(const Oxs_SimState& state) \
{ NAME##_output.cache.state_id=state.Id(); \
  NAME##_output.cache.value=state.NAME; }

OSO_FUNC(iteration_count)
OSO_FUNC(stage_iteration_count)
OSO_FUNC(stage_number)

void
Oxs_Driver::Fill__spin_output(const Oxs_SimState& state)
{
  spin_output.cache.state_id=0;
  spin_output.cache.value = state.spin;
  spin_output.cache.state_id=state.Id();
}

void
Oxs_Driver::Fill__magnetization_output(const Oxs_SimState& state)
{
  magnetization_output.cache.state_id=0;
  magnetization_output.cache.value.AdjustSize(state.mesh);
  UINT4m size=state.mesh->Size();
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& sMs = *(state.Ms);
  Oxs_MeshValue<ThreeVector>& mag = magnetization_output.cache.value;
  for(UINT4m i=0;i<size;i++) {
    mag[i] = spin[i];
    mag[i] *= sMs[i];
  }
  magnetization_output.cache.state_id=state.Id();
}

void Oxs_Driver::UpdateSpinAngleData(const Oxs_SimState& state) const
{
  REAL8m maxang,stage_maxang,run_maxang;
  maxang = stage_maxang = run_maxang = -1.0; // Safety init
  if(state.mesh->Size()<2) maxang = 0.0; // Special case
  else {
    REAL8m mindot = state.mesh->MinNeighborDot(state.spin,*(state.Ms));
    // Handle extremal cases separately in case spins drift off S2
    if(mindot >=  1)      maxang = 0.0;
    else if(mindot <= -1) maxang = 180.0;
    else                  maxang = acos(mindot)*(180.0/PI);
  }
  state.GetDerivedData("PrevState Stage Max Spin Ang",stage_maxang);
  state.GetDerivedData("PrevState Run Max Spin Ang",run_maxang);
  if(maxang>stage_maxang) stage_maxang=maxang;
  if(maxang>run_maxang)   run_maxang=maxang;
  state.AddDerivedData("Max Spin Ang",maxang);
  state.AddDerivedData("Stage Max Spin Ang",stage_maxang);
  state.AddDerivedData("Run Max Spin Ang",run_maxang);
}

void
Oxs_Driver::Fill__maxSpinAng_output(const Oxs_SimState& state) 
{
  maxSpinAng_output.cache.state_id =
  stage_maxSpinAng_output.cache.state_id =
  run_maxSpinAng_output.cache.state_id = 0;

  REAL8m maxang,stage_maxang,run_maxang;
  maxang = stage_maxang = run_maxang = -1.0; // Safety init

  if(!state.GetDerivedData("Max Spin Ang",maxang) ||
     !state.GetDerivedData("Stage Max Spin Ang",stage_maxang) ||
     !state.GetDerivedData("Run Max Spin Ang",run_maxang)) {
    UpdateSpinAngleData(state);
    state.GetDerivedData("Max Spin Ang",maxang);
    state.GetDerivedData("Stage Max Spin Ang",stage_maxang);
    state.GetDerivedData("Run Max Spin Ang",run_maxang);
  }
  maxSpinAng_output.cache.value=maxang;
  stage_maxSpinAng_output.cache.value=stage_maxang;
  run_maxSpinAng_output.cache.value=run_maxang;
  maxSpinAng_output.cache.state_id =
  stage_maxSpinAng_output.cache.state_id =
  run_maxSpinAng_output.cache.state_id = state.Id();
}

void
Oxs_Driver::Fill__aveM_output(const Oxs_SimState& state)
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& sMs = *(state.Ms);
  UINT4m size=state.mesh->Size();

  if(aveMx_output.GetCacheRequestCount()>0 &&
     aveMy_output.GetCacheRequestCount()>0 &&
     aveMz_output.GetCacheRequestCount()>0) {
    // Usual case: All three components desired
    aveMx_output.cache.state_id=0;
    aveMy_output.cache.state_id=0;
    aveMz_output.cache.state_id=0;
    REAL8m Mx=0.0;
    REAL8m My=0.0;
    REAL8m Mz=0.0;
    if(size>0) {
      UINT4m i=0;
      do {
	REAL8m sat_mag = sMs[i];
	Mx += sat_mag*(spin[i].x);
	My += sat_mag*(spin[i].y);
	Mz += sat_mag*(spin[i].z);
	++i;
      } while(i<size);
      Mx/=size;
      My/=size;
      Mz/=size;
    }
    aveMx_output.cache.value=Mx;
    aveMx_output.cache.state_id=state.Id();
    aveMy_output.cache.value=My;
    aveMy_output.cache.state_id=state.Id();
    aveMz_output.cache.value=Mz;
    aveMz_output.cache.state_id=state.Id();
  } else {
    // Calculate components on a case-by-case basis
    if(aveMx_output.GetCacheRequestCount()>0) {
      aveMx_output.cache.state_id=0;
      REAL8m Mx=0.0;
      if(size>0) {
	UINT4m i=0;
	do {
	  Mx += sMs[i]*(spin[i].x);
	  ++i;
	} while(i<size);
	Mx/=size;
      }
      aveMx_output.cache.value=Mx;
      aveMx_output.cache.state_id=state.Id();
    }

    if(aveMy_output.GetCacheRequestCount()>0) {
      aveMy_output.cache.state_id=0;
      REAL8m My=0.0;
      if(size>0) {
	UINT4m i=0;
	do {
	  My += sMs[i]*(spin[i].y);
	  ++i;
	} while(i<size);
	My/=size;
      }
      aveMy_output.cache.value=My;
      aveMy_output.cache.state_id=state.Id();
    }

    if(aveMz_output.GetCacheRequestCount()>0) {
      aveMz_output.cache.state_id=0;
      REAL8m Mz=0.0;
      if(size>0) {
	UINT4m i=0;
	do {
	  Mz += sMs[i]*(spin[i].z);
	  ++i;
	} while(i<size);
	Mz/=size;
      }
      aveMz_output.cache.value=Mz;
      aveMz_output.cache.state_id=state.Id();
    }
  }
}
