/* FILE: driver.h                 -*-Mode: c++-*-
 *
 * Abstract driver class; fills states and steps evolvers.
 *
 */

#ifndef _OXS_DRIVER
#define _OXS_DRIVER

#include "ext.h"
#include "labelvalue.h"
#include "simstate.h"
#include "key.h"
#include "output.h"

/* End includes */

class Oxs_Evolver; // Forward references
struct OxsRunEvent;

class Oxs_Driver: public Oxs_Ext {
private:
  Oxs_LabelValue attributes; // basename, etc.

  Oxs_ConstKey<Oxs_SimState> current_state;

  // State-based outputs, maintained by the driver.  These are
  // conceptually public, but are specified private to force
  // clients to use the output_map interface in Oxs_Director.
#define OSO_DECL(name) \
void Fill__##name##_output(const Oxs_SimState&); \
Oxs_ScalarOutput<Oxs_Driver> name##_output
  OSO_DECL(iteration_count);
  OSO_DECL(stage_iteration_count);
  OSO_DECL(stage_number);
  Oxs_VectorFieldOutput<Oxs_Driver> spin_output;
  void Fill__spin_output(const Oxs_SimState&);
  Oxs_VectorFieldOutput<Oxs_Driver> magnetization_output;
  void Fill__magnetization_output(const Oxs_SimState&);
  Oxs_ScalarOutput<Oxs_Driver> aveMx_output;
  Oxs_ScalarOutput<Oxs_Driver> aveMy_output;
  Oxs_ScalarOutput<Oxs_Driver> aveMz_output;
  void Fill__aveM_output(const Oxs_SimState&);
  Oxs_ScalarOutput<Oxs_Driver> maxSpinAng_output;
  Oxs_ScalarOutput<Oxs_Driver> stage_maxSpinAng_output;
  Oxs_ScalarOutput<Oxs_Driver> run_maxSpinAng_output;
  void UpdateSpinAngleData(const Oxs_SimState& state) const;
  void Fill__maxSpinAng_output(const Oxs_SimState&);

#undef OSO_DECL

  vector<UINT4m> stage_iteration_limit; // Counts; 0 => no limit
  UINT4m total_iteration_limit; // 0 => no limit
  UINT4m stage_count_request;   // 0 => unlimited
  UINT4m number_of_stages;      // 0 => unlimited
  /// stage_count_request is set in Specify block via
  /// stage_count parameter.  number_of_stages is set
  /// via call to Oxs_Director::ExtObjStageRequestCounts().

  INT4m stage_count_check; // 1 => raise error if
  /// stage counts across Oxs_Ext objects aren't compatible.


  // Routines defined by child classes to detect stage and
  // problem end events.
  virtual BOOL ChildIsStageDone(const Oxs_SimState& state) const =0;
  virtual BOOL ChildIsRunDone(const Oxs_SimState& state) const =0;

protected:
  Oxs_Driver(const char* name,        // Child instance id
             Oxs_Director* newdtr,    // App director
	     const char* argstr);     // Args

  BOOL Init();  // All children of Oxs_Driver *must* call
  /// this function in their Init() routines.  The main purpose
  /// of this function is to initialize the current state.

public:

  virtual ~Oxs_Driver();

  virtual void StageRequestCount(unsigned int& min,
				 unsigned int& max) const;
  // Number of stages wanted by driver

  virtual Oxs_ConstKey<Oxs_SimState> GetInitialState() const =0;

  const Oxs_SimState* GetCurrentState() {
    return current_state.GetPtr();
  }

  // Checks state against parent Oxs_Driver class stage/run limiters,
  // and if passes (i.e., not done), then calls ChildStage/RunDone
  // functions.  If this is not enough flexibility, we could make
  // these functions virtual, and move them into the protected:
  // or public: sections, to allow child overrides.
  BOOL IsStageDone(const Oxs_SimState& state) const;
  BOOL IsRunDone(const Oxs_SimState& state) const;

  virtual  BOOL
  Step(Oxs_ConstKey<Oxs_SimState> current_state,
       Oxs_Key<Oxs_SimState>& next_state)=0;
  // Returns true if step was successful, false if
  // unable to step as requested.

  // Problem "Run" interface; called from director
  void Run(vector<OxsRunEvent>& results);

  // SetStage and GetStage throw exceptions on errors.
  void SetStage(UINT4m requestedStage);
  UINT4m GetStage() const;
  UINT4m GetNumberOfStages() const { return number_of_stages; }

  virtual void FillState(const Oxs_SimState& old_state,
                         Oxs_SimState& new_state) const;
  /// FillState is called to fill in a new blank state, using
  ///   an old state as a template.  Specific drivers may have
  ///   supplemental state filling routines as part of a cooperative
  ///   agreement with particular evolver subclasses.  A default
  ///   method is provided.  Child classes should either use this
  ///   method as is, or else call it from overriding class.
  ///   Complete replacement is discouraged, to make it easier
  ///   to propagate any future changes in simstate structure.

  virtual void FillNewStageState(const Oxs_SimState& old_state,
				 int new_stage_number,
				 Oxs_SimState& new_state) const;
  /// FillNewStageState is called to copy a state across a stage
  /// boundary.  A default method is provided.  Child classes should
  /// either use this method as is, or else call it from overriding
  /// class.  Complete replacement is discouraged, to make it easier to
  /// propagate any future changes in simstate structure.

  // Attribute access.  The following 3 labels are guaranteed
  // to exist: basename, vector_field_output_format, and
  // scalar_output_format.
  void GetAttributeLabels(vector<string>& labels) const {
    attributes.GetLabels(labels);
  }
  string GetAttribute(const string& label) const {
    return attributes.GetValue(label);
  } // Throws an exception if label is not in attribute list.

};

#endif // _OXS_DRIVER
