/* FILE: oxscmd.cc                 -*-Mode: c++-*-
 *
 * Tcl interface to C/C++ code in oxs.
 *
 * Each routine in the file is registered with the Tcl interpreter
 * by OxsRegisterInterfaceCommands() at the bottom of this file.
 * OxsRegisterInterfaceCommands() is called by Oxs_Init().
 * (Note: OxsRegisterInterfaceCommands() is not in any header file
 * because it should only be called by Oxs_Init().)
 *
 */

#include <vector>

#include "director.h"
#include "driver.h"  // For Oxs_Driver::RunEvent structure
#include "ext.h"
#include "oc.h"
#include "output.h"
#include "energy.h"	// Needed to make MSVC++ 5 happy

/* End includes */     

Tcl_CmdProc	Oxs_ProbInit;
Tcl_CmdProc	Oxs_ProbReset;
Tcl_CmdProc	Oxs_Run;
Tcl_CmdProc	Oxs_ProbRelease;
Tcl_CmdProc	Oxs_Driver;
Tcl_CmdProc	Oxs_ListRegisteredExtClasses;
Tcl_CmdProc	Oxs_ListEnergyObjects;
Tcl_CmdProc	Oxs_ListExtObjects;
Tcl_CmdProc	Oxs_ListOutputObjects;
Tcl_CmdProc	Oxs_OutputGet;
Tcl_CmdProc	Oxs_OutputNames;
Tcl_CmdProc	Oxs_SetStage;
Tcl_CmdProc	Oxs_GetStage;
Tcl_CmdProc	Oxs_GetNumberOfStages;
Tcl_CmdProc	Oxs_GetMif;

/*
 *----------------------------------------------------------------------
 *
 * Program globals
 *
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 *
 * Oxs_SetStage --
 *
 *      Interactively set the Stage of the problem being solved.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int Oxs_SetStage(ClientData cd,Tcl_Interp *interp,
			    int argc,CONST84 char **argv)
{
  int requestedStage,errcode;
  string errmsg;
  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # of args: should be \"", argv[0],
		     "stage_number\"", (char *) NULL);
    return TCL_ERROR;
  }
  if (Tcl_GetInt(interp, argv[1], &requestedStage) != TCL_OK) {
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  errcode = director->SetStage(requestedStage, errmsg);
  if (errcode!=TCL_OK) {
    Tcl_AppendResult(interp,errmsg.c_str(),(char *)NULL);
  }
  return errcode;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_GetStage --
 *
 *      Interactively get the Stage of the problem being solved.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int Oxs_GetStage(ClientData cd,Tcl_Interp *interp,
			    int argc,CONST84 char **argv)
{
  int stage,errcode;
  string errmsg;
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be \"", argv[0],
		     "\"", (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  errcode = director->GetStage(stage, errmsg);
  if (errcode!=TCL_OK) {
    Tcl_AppendResult(interp,errmsg.c_str(),(char *)NULL);
  } else {
    char buf[64];
    Oc_Snprintf(buf,sizeof(buf),"%d",stage);
    Tcl_AppendResult(interp,buf,(char *) NULL);
  }
  return errcode;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_GetMif --
 *
 *      Interactively get the MIF handle of the problem being solved.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int Oxs_GetMif(ClientData cd,Tcl_Interp *interp,
			    int argc, CONST84 char **argv)
{
  int errcode;
  string mif;
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be \"", argv[0],
		     "\"", (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  errcode = director->GetMif(mif);
  Tcl_AppendResult(interp,mif.c_str(),(char *)NULL);
  return errcode;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_GetNumberOfStages --
 *
 *      Retrieve the number of stages in the problem being solved.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int Oxs_GetNumberOfStages(ClientData cd,Tcl_Interp *interp,
			  int argc,CONST84 char **argv)
{
  int stage_count,errcode;
  string errmsg;
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be \"", argv[0],
		     "\"", (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  errcode = director->GetNumberOfStages(stage_count, errmsg);
  if (errcode!=TCL_OK) {
    Tcl_AppendResult(interp,errmsg.c_str(),(char *)NULL);
  } else {
    char buf[64];
    Oc_Snprintf(buf,sizeof(buf),"%d",stage_count);
    Tcl_AppendResult(interp,buf,(char *) NULL);
  }
  return errcode;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbInit --
 *
 *      Micromagnetic problem initialization
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int Oxs_ProbInit(ClientData cd,Tcl_Interp *interp,
			    int argc,CONST84 char **argv)
{
  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # of args: should be",
		     " \"Oxs_ProbInit filename\"",
		     (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  int result=TCL_ERROR;
  try {
    result = director->ProbInit(argv[1]);
  } catch (Oc_Exception& err) {
    Tcl_AppendResult(interp,"Error thrown from inside"
		     " director ProbInit function---",
		     err.MessageType(),"---",
		     err.MessageText().c_str(),
		     (char *) NULL);
    result = TCL_ERROR; // Safety
  } catch (EXCEPTION& err) {
    Tcl_AppendResult(interp,"Library error thrown from inside"
		     " director ProbInit function---",
		     err.what(),(char *) NULL);
    result = TCL_ERROR; // Safety
#ifndef NDEBUG
  } catch (...) {
    Tcl_AppendResult(interp,"Unrecognized error thrown from inside"
		     " director ProbInit function.",
		     (char *) NULL);
    result = TCL_ERROR; // Safety
#endif // NDEBUG
  }
  return result;
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbReset --
 *
 *      Resets micromagnetic problem
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_ProbReset(ClientData cd,Tcl_Interp *interp,
			     int argc,CONST84 char**)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be just",
		     " \"Oxs_ProbReset\"",
		     (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  int result = TCL_ERROR;
  try {
    result = director->ProbReset();
  } catch (Oc_Exception& err) {
    Tcl_AppendResult(interp,"Error thrown from inside"
		     " director ProbReset function---",
		     err.MessageType(),"---",
		     err.MessageText().c_str(),
		     (char *) NULL);
    result = TCL_ERROR; // Safety
  } catch (EXCEPTION& err) {
    Tcl_AppendResult(interp,"System library error thrown from inside"
		     " director ProbReset function---",
		     err.what(),(char *) NULL);
    result = TCL_ERROR; // Safety
#ifndef NDEBUG
  } catch (...) {
    Tcl_AppendResult(interp,"Unrecognized error thrown from inside"
		     " director ProbReset function.",
		     (char *) NULL);
    result = TCL_ERROR; // Safety
#endif // NDEBUG
  }

  return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_Run --
 *
 *      Runs micromagnetic problem
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_Run(ClientData cd,Tcl_Interp *interp,
	    int argc,CONST84 char** /* argv */)
{ // On success, returns a list of events with associated state id
  // info.  Each item in the list has the form
  //      <event_type state_id stage_number iteration_count>
  // In the future, we might want to change this to return instead
  // some general state reference that the script can query for
  // itself to pull out the info that it needs.
  // NOTE: Each element of the "results" vector holds a read lock
  //  on the associated state, until such time as "results" is
  //  destroyed.
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be just",
		     " \"Oxs_Run\"",
		     (char *) NULL);
    return TCL_ERROR;
  }

  Oxs_Director* director = (Oxs_Director *)cd;

  string errmsg;
  vector<OxsRunEvent> results;
  int errcode = director->ProbRun(results,errmsg);
  if(errcode==0 && results.size()>0) {
    // Build event list
    vector<string> event_list;
    vector<OxsRunEvent>::const_iterator it;
    for(it=results.begin();it!=results.end();++it) {
      const Oxs_SimState* state = it->state_key.GetPtr();
      if(state==NULL || state->Id()==0) {
	Tcl_AppendResult(interp,"Invalid state returned",(char *) NULL);
	return TCL_ERROR;
      }
      static char event_buf[512];
      switch(it->event) {
      case OXS_STEP_EVENT:
	Oc_Snprintf(event_buf,sizeof(event_buf),
		    "STEP %u %u %u",
		    state->Id(),
		    state->stage_number,
		    state->iteration_count);
	break;
      case OXS_STAGE_DONE_EVENT:
	Oc_Snprintf(event_buf,sizeof(event_buf),
		    "STAGE_DONE %u %u %u",
		    state->Id(),
		    state->stage_number,
		    state->iteration_count);
	break;
      case OXS_RUN_DONE_EVENT:
	Oc_Snprintf(event_buf,sizeof(event_buf),
		    "RUN_DONE %u %u %u",
		    state->Id(),
		    state->stage_number,
		    state->iteration_count);
	break;
      default:
	char buf[1024];
	Oc_Snprintf(buf,sizeof(buf),
		    "Unknown event type returned: %d",
		    static_cast<int>(it->event));
	Tcl_AppendResult(interp,buf,(char *) NULL);
	return TCL_ERROR;
      }
      event_list.push_back(string(event_buf));
    }
    string whole_list = Oxs_MergeList(event_list);
    Tcl_AppendResult(interp,whole_list.c_str(),(char *) NULL);
  }

  if(errcode!=0 && errmsg.length()>0) {
    Tcl_AppendResult(interp,errmsg.c_str(),(char *) NULL);
  }

  if(errcode!=0) return TCL_ERROR;
  return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbRelease --
 *
 *      Releases micromagnetic problem statement.  Depending on OS,
 *      this may free some memory, but is intended mainly to put
 *      the solver in a well-defined state.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_ProbRelease(ClientData cd,Tcl_Interp *interp,
			     int argc,CONST84 char**)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be just",
		     " \"Oxs_ProbRelease\"",
		     (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;

  int result = TCL_ERROR;
  try {
    result = director->ProbRelease();
  } catch (Oc_Exception& err) {
    Tcl_AppendResult(interp,"Error thrown from inside"
		     " director ProbRelease function---",
		     err.MessageType(),"---",
		     err.MessageText().c_str(),
		     (char *) NULL);
    result = TCL_ERROR; // Safety
  } catch (EXCEPTION& err) {
    Tcl_AppendResult(interp,"Library error thrown from inside"
		     " director ProbRelease function---",
		     err.what(),(char *) NULL);
    result = TCL_ERROR; // Safety
#ifndef NDEBUG
  } catch (...) {
    Tcl_AppendResult(interp,"Unrecognized error thrown from inside"
		     " director ProbRelease function.",
		     (char *) NULL);
    result = TCL_ERROR; // Safety
#endif // NDEBUG
  }

  return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_Driver --
 *
 *      Returns properties of the driver.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_Driver(ClientData cd,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  BOOL bad_param_count=0;
  BOOL bad_driver_init=0;
  Oxs_Director* director = (Oxs_Director *)cd;

  if (argc < 2) {
    bad_param_count=1;
  } else if(strcmp("name",argv[1])==0) {
    if(argc!=2) {
      bad_param_count=1;
    } else {
      string name;
      if(director->GetDriverInstanceName(name)!=0) {
	bad_driver_init = 1;
      } else {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp,name.c_str(),(char *)NULL);
      }
    }
  } else if(strcmp("attributes",argv[1])==0) {
    if(argc!=2) {
      bad_param_count=1;
    } else {
      vector<string> labels;
      if(director->GetDriverAttributes(labels) != 0) {
	bad_driver_init = 1;
      } else {
	Tcl_ResetResult(interp);
	vector<string>::const_iterator it = labels.begin();
	while(it != labels.end()) {
	  Tcl_AppendElement(interp, Oc_AutoBuf(it->c_str()));
	  ++it;
	}
      }
    }
  } else if(strcmp("value",argv[1])==0) {
    if(argc!=3) {
      bad_param_count=1;
    } else {
      string value;
      int errcode
	= director->GetDriverAttributeValue(string(argv[2]),value);
      if(errcode==0) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp,value.c_str(),(char *)NULL);
      } else if(errcode==1) {
	bad_driver_init = 1;
      } else {
	Tcl_AppendResult(interp, "Label \"", argv[2],
			 "\" not recognized by driver.",
			 (char *) NULL);
	return TCL_ERROR;
      }
    }
  } else {
    Tcl_AppendResult(interp, "bad option \"",argv[1],
		     "\": must be name, attributes or value",
		     (char *) NULL);
    return TCL_ERROR;
  }

  if(bad_param_count) {
    Tcl_AppendResult(interp, "wrong # of args: should be \"",
		     argv[0]," option ?arg?\"",(char *) NULL);
    return TCL_ERROR;
  }

  if(bad_driver_init) {
    Tcl_AppendResult(interp, "Driver not initialized.",
		     (char *) NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListRegisteredExtClasses --
 *
 *      Lists Ext child classes that are loaded and registered
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_ListRegisteredExtClasses(ClientData,
				 Tcl_Interp *interp,
				 int,CONST84 char**)
{
  string result=Oxs_Ext::ListRegisteredChildClasses();

  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,result.c_str(),(char *)NULL);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListExtObjects --
 *
 *      Lists Ext objects that have been created
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_ListExtObjects(ClientData cd,Tcl_Interp *interp,
		       int,CONST84 char**)
{
  Oxs_Director* director = (Oxs_Director *)cd;
  string result=director->ListExtObjects();
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,result.c_str(),(char *)NULL);
  return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListEnergyObjects --
 *
 *      Lists Energy objects that have been created
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_ListEnergyObjects(ClientData cd,Tcl_Interp *interp,
			  int,CONST84 char**)
{
  Oxs_Director* director = (Oxs_Director *)cd;
  string result=director->ListEnergyObjects();
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,result.c_str(),(char *)NULL);
  return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListOutputObjects --
 *
 *      Lists all output objects that are registered with the director.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_ListOutputObjects(ClientData cd, Tcl_Interp *interp,
			  int argc, CONST84 char **argv)
{
    if (argc > 1) {
        Tcl_AppendResult(interp, "wrong # of args: should be \"",
		     argv[0], "\"", (char *) NULL);
	return TCL_ERROR;
    }

    Oxs_Director* director = (Oxs_Director *)cd;

    vector<string> outputs;
    director->ListOutputObjects(outputs);

    Tcl_ResetResult(interp);
    vector<string>::const_iterator it = outputs.begin();
    while(it != outputs.end()) {
	Tcl_AppendElement(interp, Oc_AutoBuf(it->c_str()));
	++it;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_OutputGet--
 *
 *      Calls output method of director.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_OutputGet(ClientData cd, Tcl_Interp *interp, int argc, CONST84 char **argv)
{
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # of args: should be \"",
		     argv[0], " output ...\"", (char *) NULL);
	return TCL_ERROR;
    }
    Oxs_Director *director = (Oxs_Director *) cd;
    return director->Output(argv[1], interp, argc-2, argv+2);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_OutputNames --
 *
 *      Retrieves names that describe an output.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      Sets result of interp to a triple of strings giving the
 *      source, name, and type of output.
 *
 *----------------------------------------------------------------------
 */
int Oxs_OutputNames(ClientData cd, Tcl_Interp *interp, int argc, CONST84 char **argv)
{
    if (argc != 2) {
        Tcl_AppendResult(interp, "wrong # of args: should be \"",
		     argv[0], " output\"", (char *) NULL);
	return TCL_ERROR;
    }
    Oxs_Director *director = (Oxs_Director *) cd;

#ifdef OXS_OUTPUTNAMECOUNT
#error Whoops, OXS_OUTPUTNAMECOUNT macro already defined
#endif
#define OXS_OUTPUTNAMECOUNT 4
    string ownerName, outputName, outputType, outputUnits;
    char *v[OXS_OUTPUTNAMECOUNT];
    director->OutputNames(argv[1],
			  ownerName, outputName,
			  outputType, outputUnits);
    v[0]=const_cast<char*>(ownerName.c_str());
    v[1]=const_cast<char*>(outputName.c_str());
    v[2]=const_cast<char*>(outputType.c_str());
    v[3]=const_cast<char*>(outputUnits.c_str());
    Tcl_ResetResult(interp);
    char* namelist = Tcl_Merge(OXS_OUTPUTNAMECOUNT,v);
    Tcl_AppendResult(interp, namelist, (char *) NULL);
    Tcl_Free(namelist);
    return TCL_OK;
#undef OXS_OUTPUTNAMECOUNT
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_DirectorDevelopTest --
 *
 * Hook for development/debug testing.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
Tcl_CmdProc	Oxs_DirectorDevelopTest;
int Oxs_DirectorDevelopTest(ClientData cd,Tcl_Interp *interp,
			    int argc,CONST84 char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    Tcl_AppendResult(interp, "wrong # of args: should be \"",
		     argv[0],"count\"",(char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  string out;
  int result_code = director->DoDevelopTest(string(argv[1]),out);
  if(result_code!=0) {
    char buf[256];
    Oc_Snprintf(buf,sizeof(buf),"%d",result_code);
    Tcl_AppendResult(interp,
		     "Test error; error code: ",buf,
		     "\nSee DoDevelopTest source code in"
		     " oommf/app/oxs/base/director.cc"
		     " for details.",
		     (char *)NULL);
    return TCL_ERROR;
  }
  Tcl_AppendResult(interp,out.c_str(),(char *)NULL);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * OxsRegisterInterfaceCommands --
 *
 *      Registers each of the above routines with the Tcl interpreter
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
void OxsRegisterInterfaceCommands(Oxs_Director* director,
				  Tcl_Interp *interp)
{
#define REGCMD(foo) \
  Tcl_CreateCommand(interp,Oc_AutoBuf(#foo),foo,director,NULL)
  REGCMD(Oxs_ProbInit);
  REGCMD(Oxs_ProbReset);
  REGCMD(Oxs_Run);
  REGCMD(Oxs_ProbRelease);
  REGCMD(Oxs_Driver);
  REGCMD(Oxs_ListRegisteredExtClasses);
  REGCMD(Oxs_ListEnergyObjects);
  REGCMD(Oxs_ListExtObjects);
  REGCMD(Oxs_ListOutputObjects);
  REGCMD(Oxs_OutputGet);
  REGCMD(Oxs_OutputNames);
  REGCMD(Oxs_SetStage);
  REGCMD(Oxs_GetStage);
  REGCMD(Oxs_GetNumberOfStages);
  REGCMD(Oxs_DirectorDevelopTest);
  REGCMD(Oxs_GetMif);

#undef REGCMD
}

