/* FILE: cubicanisotropy.cc            -*-Mode: c++-*-
 *
 * Cubic Anisotropy, derived from Oxs_Energy class.
 *
 */

#include "oc.h"
#include "nb.h"
#include "threevector.h"
#include "director.h"
#include "simstate.h"
#include "ext.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "cubicanisotropy.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_CubicAnisotropy);

/* End includes */


// Constructor
Oxs_CubicAnisotropy::Oxs_CubicAnisotropy(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr), mesh_id(0)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("K1",Oxs_ScalarField,K1_init);
  OXS_GET_INIT_EXT_OBJECT("axis1",Oxs_VectorField,axis1_init);
  OXS_GET_INIT_EXT_OBJECT("axis2",Oxs_VectorField,axis2_init);
  VerifyAllInitArgsUsed();
}

void Oxs_CubicAnisotropy::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  UINT4m size = state.mesh->Size();
  if(mesh_id != state.mesh->Id()) {
    // This is either the first pass through, or else mesh
    // has changed.
    mesh_id=0;
    K1_init->FillMeshValue(state.mesh,K1);
    axis1_init->FillMeshValue(state.mesh,axis1);
    axis2_init->FillMeshValueOrthogonal(state.mesh,axis1,axis2);
    for(UINT4m i=0;i<size;i++) {
      // Much of the code below requires axis1 and axis2 to be
      // orthogonal unit vectors.  Guarantee this is the case:
      const REAL8m eps = 1e-14;
      if(fabs(axis1[i].MagSq()-1)>eps) {
	string msg="Invalid initialization detected for object "
	  + string(InstanceName())
	  + ": Anisotropy axis 1 isn't norm 1";
	throw Oxs_Ext::Error(msg.c_str());
      }
      if(fabs(axis2[i].MagSq()-1)>eps) {
	string msg="Invalid initialization detected for object "
	  + string(InstanceName())
	  + ": Anisotropy axis 2 isn't norm 1";
	throw Oxs_Ext::Error(msg.c_str());
      }
      if(fabs(axis1[i]*axis2[i])>eps) {
	string msg="Invalid initialization detected for object "
	  + string(InstanceName())
	  + ": Specified anisotropy axes aren't perpendicular";
	throw Oxs_Ext::Error(msg.c_str());
      }
    }
    mesh_id=state.mesh->Id();
  }


  for(UINT4m i=0;i<size;++i) {
    const ThreeVector& u1 = axis1[i];
    const ThreeVector& u2 = axis2[i];
    const ThreeVector&  m = spin[i];

    // This code requires u1 and u2 to be orthonormal, and m to be a
    // unit vector.  Basically, decompose
    //
    //             m = a1.u1 + a2.u2 + a3.u3
    //               =  m1   +  m2   +  m3
    //
    // where u3=u1xu2, a1=m*u1, a2=m*u2, a3=m*u3.
    //
    // Then the energy is
    //                 2  2     2  2     2  2
    //            K (a1 a2  + a1 a3  + a2 a3 )
    //
    // and the field in say the u1 direction is
    //              2    2                 2    2
    //         C (a2 + a3 ) a1 . u1 = C (a2 + a3 ) m1
    //
    // where C = -2K/(MU0 Ms).
    //
    // In particular, note that
    //           2         2     2
    //         a3  = 1 - a1  - a2
    // and
    //         m3  = m - m1 - m2
    //
    // This shows that energy and field can be computed without
    // explicitly calculating u3.  However, the cross product
    // evaluation to get u3 is not that expensive, and appears
    // to be more accurate.  At a minimum, in the above expressions
    // one should at least insure that a3^2 is non-negative.

    REAL8m k = K1[i];
    REAL8m field_mult = (-2/MU0)*k*Ms_inverse[i];
    if(field_mult==0.0) {
      energy[i]=0.0;
      field[i].Set(0.,0.,0.);
      continue;
    }

    ThreeVector u3 = u1;    u3 ^= u2;
    REAL8m a1 = u1*m;  REAL8m a1sq = a1*a1;
    REAL8m a2 = u2*m;  REAL8m a2sq = a2*a2;
    REAL8m a3 = u3*m;  REAL8m a3sq = a3*a3;

    energy[i] = k * (a1sq*a2sq+a1sq*a3sq+a2sq*a3sq);

    ThreeVector m1 = a1*u1;
    ThreeVector m2 = a2*u2;
    ThreeVector m3 = a3*u3;
    field[i]  = (a2sq+a3sq)*m1;
    field[i] += (a1sq+a3sq)*m2;
    field[i] += (a1sq+a2sq)*m3;
    field[i] *= field_mult;
  }
}
