/* FILE: demag.cc            -*-Mode: c++-*-
 *
 * Average H demag field across rectangular cells.  This is a modified
 * version of the simpledemag class, which uses symmetries in the
 * interaction coefficients to reduce memory usage.
 *
 */

#include "nb.h"
#include "director.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

#include "rectangularmesh.h"
#include "demag.h"
#include "demagcoef.h"
#include "fft.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_Demag);

/* End includes */

// Helper function
UINT4m Oxs_Demag::NextPowerOfTwo(UINT4m n) const
{ // Returns first power of two >= n
  UINT4m m=1;
  while(m<n) m*=2;
  return m;
}

// Constructor
Oxs_Demag::Oxs_Demag(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr),
    rdimx(0),rdimy(0),rdimz(0),
    cdimx(0),cdimy(0),cdimz(0),cstridey(0),cstridez(0),
    mesh_id(0),
    adimx(0),adimy(0),adimz(0),astridey(0),astridez(0),
    A00(NULL),A01(NULL),A02(NULL),A11(NULL),A12(NULL),A22(NULL),
    xcomp(NULL),ycomp(NULL),zcomp(NULL)
{
  kernel_fill_routine = GetStringInitValue("kernel_fill_routine","new");
  VerifyAllInitArgsUsed();
}

void Oxs_Demag::ReleaseMemory()
{
  if(xcomp!=NULL) { delete[] xcomp; xcomp=NULL; }
  if(ycomp!=NULL) { delete[] ycomp; ycomp=NULL; }
  if(zcomp!=NULL) { delete[] zcomp; zcomp=NULL; }
  if(A00!=NULL) { delete[] A00; A00=NULL; }
  if(A01!=NULL) { delete[] A01; A01=NULL; }
  if(A02!=NULL) { delete[] A02; A02=NULL; }
  if(A11!=NULL) { delete[] A11; A11=NULL; }
  if(A12!=NULL) { delete[] A12; A12=NULL; }
  if(A22!=NULL) { delete[] A22; A22=NULL; }
  rdimx=rdimy=rdimz=0;
  cdimx=cdimy=cdimz=0;
  cstridey=cstridez=0;
  adimx=adimy=adimz=0;
  astridey=astridez=0;
}

void Oxs_Demag::FillCoefficientArrays(const Oxs_Mesh* genmesh) const
{ // This routine is conceptually const.

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(genmesh);
  if(mesh==NULL) {
    string msg=string("Object ")
      + string(genmesh->InstanceName())
      + string(" is not a rectangular mesh.");
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Fill dimension variables
  rdimx = mesh->DimX();
  rdimy = mesh->DimY();
  rdimz = mesh->DimZ();
  if(rdimx==0 || rdimy==0 || rdimz==0) return; // Empty mesh!

  cdimx = NextPowerOfTwo(2*rdimx)/2;
  if(rdimy>1) cdimy = NextPowerOfTwo(2*rdimy);
  else        cdimy = 1;
  if(rdimz>1) cdimz = NextPowerOfTwo(2*rdimz);
  else        cdimz = 1;
  cstridey = cdimx + 1;  // Pad by one to avoid cache line entanglement
  cstridez = cstridey*cdimy;

  UINT4m ctotalsize=cstridez*cdimz;
  UINT4m rtotalsize=2*ctotalsize;
  if(rtotalsize<2*cdimx || rtotalsize<cdimy || rtotalsize<cdimz) {
    // Partial overflow check
    string msg="UINT4m overflow in " + string(InstanceName())
      + ": Product cdimx*cdimy*cdimz too big to fit in a UINT4m variable";
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Allocate memory for interaction matrices and magnetization components
  xcomp = new Oxs_Complex[ctotalsize];
  ycomp = new Oxs_Complex[ctotalsize];
  zcomp = new Oxs_Complex[ctotalsize];
  adimx=1+cdimx;
  adimy=1+cdimy/2;
  adimz=1+cdimz/2;
  astridey=adimx;
  astridez=adimy*astridey;
  UINT4m atotalsize=adimz*astridez;
  A00   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A01   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A02   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A11   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A12   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A22   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  if(xcomp==NULL || ycomp==NULL || zcomp==NULL
     || A00==NULL || A01==NULL || A02==NULL
     || A11==NULL || A12==NULL || A22==NULL) {
    // Safety check for those machines on which new[] doesn't throw
    // BadAlloc.
    string msg="Insufficient memory in Demag constructor.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  REALWIDE *rxcomp = static_cast<REALWIDE*>(static_cast<void*>(xcomp));
  REALWIDE *rycomp = static_cast<REALWIDE*>(static_cast<void*>(ycomp));
  REALWIDE *rzcomp = static_cast<REALWIDE*>(static_cast<void*>(zcomp));
  UINT4m rstridey=2*cstridey;
  UINT4m rstridez=2*cstridez;

  // According (16) in Newell's paper, the demag field is given by
  //                        H = -N*M
  // where N is the "demagnetizing tensor," with components Nxx, Nxy,
  // etc.  With the '-1' in 'scale' we store '-N' instead of 'N',
  // so we don't have to multiply the output from the FFT + iFFT
  // by -1 in ConstMagField() below.

  // Fill interaction matrices with demag coefs from Newell's paper.
  // Note that A00, A11 and A22 are even in x,y and z.
  // A01 is odd in x and y, even in z.
  // A02 is odd in x and z, even in y.
  // A12 is odd in y and z, even in x.
  // We use these symmetries to reduce storage requirements.  If
  // f is real and even, then f^ is also real and even.  If f
  // is real and odd, then f^ is (pure) imaginary and odd.
  // As a result, the transform of each of the Axx interaction
  // matrices will be real, with the same even/odd properties.
  //
  // Notation:  A00:=fs*Nxx, A01:=fs*Nxy, A02:=fs*Nxz,
  //                         A11:=fs*Nyy, A12:=fs*Nyz
  //                                      A22:=fs*Nzz
  //  where fs = -1/(2*cdimx*cdimy*cdimz)

  UINT4m index,i,j,k,istop,jstop,kstop;

  REALWIDE dx = mesh->EdgeLengthX();
  REALWIDE dy = mesh->EdgeLengthY();
  REALWIDE dz = mesh->EdgeLengthZ();
  // For demag calculation, all that matters is the relative
  // size of dx, dy and dz.  To help insure we don't run
  // outside floating point range, rescale these values so
  // largest is 1.0
  REALWIDE maxedge=dx;
  if(dy>maxedge) maxedge=dy;
  if(dz>maxedge) maxedge=dz;
  dx/=maxedge; dy/=maxedge; dz/=maxedge;

  REALWIDE scale = 1.0/(4*PI*dx*dy*dz);

  // Also throw in FFT scaling.  This allows the "NoScale" FFT routines
  // to be used.
  scale *= 1.0/(2*cdimx*cdimy*cdimz);

  for(index=0;index<ctotalsize;index++) {
    // Initialize temp space to zeros
    xcomp[index].Set(0.,0.);
    ycomp[index].Set(0.,0.);
    zcomp[index].Set(0.,0.);
  }

  // Calculate Nxx, Nyy and Nzz in first octant.
  // Step 1: Evaluate f at each cell site.  Offset by (-dx,-dy,-dz)
  //  so we can do 2nd derivative operations "in-place".
  kstop=1; if(rdimz>1) kstop=rdimz+2;
  jstop=1; if(rdimy>1) jstop=rdimy+2;
  istop=1; if(rdimx>1) istop=rdimx+2;
  for(k=0;k<kstop;k++) {
    UINT4m kindex = k*rstridez;
    REALWIDE z = 0.0;
    if(k==0) z = -dz;
    else if(k>1) z = dz*(k-1);
    for(j=0;j<jstop;j++) {
      UINT4m jkindex = kindex + j*rstridey;
      REALWIDE y = 0.0;
      if(j==0) y = -dy;
      else if(j>1) y = dy*(j-1);
      for(i=0;i<istop;i++) {
	index = i+jkindex;
#ifndef NDEBUG
	if(index>=rtotalsize) {
	  string msg="Programming error: array index out-of-bounds.";
	  throw Oxs_Ext::Error(msg.c_str());
	}
#endif // NDEBUG
	REALWIDE x = 0.0;
	if(i==0) x = -dx;
	else if(i>1) x = dx*(i-1);
	rxcomp[index] = scale*Newell_f(x,y,z);
	rycomp[index] = scale*Newell_f(y,x,z);
	rzcomp[index] = scale*Newell_f(z,y,x);
      }
    }
  }

  // Step 2a: Do d^2/dz^2
  if(kstop==1) {
    // Only 1 layer in z-direction of f stored in r?comp arrays.
    for(j=0;j<jstop;j++) {
      UINT4m jkindex = j*rstridey;
      REALWIDE y = 0.0;
      if(j==0) y = -dy;
      else if(j>1) y = dy*(j-1);
      for(i=0;i<istop;i++) {
	index = i+jkindex;
	REALWIDE x = 0.0;
	if(i==0) x = -dx;
	else if(i>1) x = dx*(i-1);
	rxcomp[index] += scale*(Newell_f(x,y,dz)-2*Newell_f(x,y,0));
	rycomp[index] += scale*(Newell_f(y,x,dz)-2*Newell_f(y,x,0));
	rzcomp[index] += scale*(Newell_f(dz,y,x)-2*Newell_f(0,y,x));
      }
    }
  } else {
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      for(j=0;j<jstop;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<istop;i++) {
	  index = i+jkindex;
	rxcomp[index] +=
	  -2*rxcomp[index+rstridez] + rxcomp[index+2*rstridez];
	rycomp[index] +=
	  -2*rycomp[index+rstridez] + rycomp[index+2*rstridez];
	rzcomp[index] +=
	  -2*rzcomp[index+rstridez] + rzcomp[index+2*rstridez];
	}
      }
    }
    for(k=rdimz;k<kstop;k++) { // Zero-fill overhang
      UINT4m kindex = k*rstridez;
      for(j=0;j<jstop;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] = 0.0;
	  rycomp[index] = 0.0;
	  rzcomp[index] = 0.0;
	}
      }
    }
  }
  // Step 2b: Do d^2/dy^2
  if(jstop==1) {
    // Only 1 layer in y-direction of f stored in r?comp arrays.
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      REALWIDE z = dz*k;
      for(i=0;i<istop;i++) {
	index = i+kindex;
	REALWIDE x = 0.0;
	if(i==0) x = -dx;
	else if(i>1) x = dx*(i-1);
	rxcomp[index] += scale * 
	  ((Newell_f(x,dy,z-dz)-2*Newell_f(x,dy,z)+Newell_f(x,dy,z+dz))
	   -2*(Newell_f(x,0,z-dz)-2*Newell_f(x,0,z)+Newell_f(x,0,z+dz)));
	rycomp[index] += scale * 
	  ((Newell_f(dy,x,z-dz)-2*Newell_f(dy,x,z)+Newell_f(dy,x,z+dz))
	   -2*(Newell_f(0,x,z-dz)-2*Newell_f(0,x,z)+Newell_f(0,x,z+dz)));
	rzcomp[index] += scale * 
	  ((Newell_f(z-dz,dy,x)-2*Newell_f(z,dy,x)+Newell_f(z+dz,dy,x))
	   -2*(Newell_f(z-dz,0,x)-2*Newell_f(z,0,x)+Newell_f(z+dz,0,x)));
      }
    }
  } else {
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      for(j=0;j<rdimy;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] +=
	    -2*rxcomp[index+rstridey] + rxcomp[index+2*rstridey];
	  rycomp[index] +=
	    -2*rycomp[index+rstridey] + rycomp[index+2*rstridey];
	  rzcomp[index] +=
	    -2*rzcomp[index+rstridey] + rzcomp[index+2*rstridey];
	}
      }
    }
    for(k=0;k<rdimz;k++) { // Zero-fill overhang
      UINT4m kindex = k*rstridez;
      for(j=rdimy;j<jstop;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] = 0.0;
	  rycomp[index] = 0.0;
	  rzcomp[index] = 0.0;
	}
      }
    }
  }
  // Step 2c: Do d^2/dx^2
  if(istop==1) {
    // Only 1 layer in x-direction of f stored in r?comp arrays.
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      REALWIDE z = dz*k;
      for(j=0;j<rdimy;j++) {
	index = kindex + j*rstridey;
	REALWIDE y = dy*j;
	rxcomp[index] += scale * 
	  (4*Newell_f(dx,y,z)
	   -2*(Newell_f(dx,y+dy,z)+Newell_f(dx,y-dy,z)
	       +Newell_f(dx,y,z+dz)+Newell_f(dx,y,z-dz))
	   +(Newell_f(dx,y+dy,z+dz)+Newell_f(dx,y-dy,z+dz)
	     +Newell_f(dx,y+dy,z-dz)+Newell_f(dx,y-dy,z-dz))
	   -8*Newell_f(0,y,z)
	   +4*(Newell_f(0,y+dy,z)+Newell_f(0,y-dy,z)
	       +Newell_f(0,y,z+dz)+Newell_f(0,y,z-dz))
	   -2*(Newell_f(0,y+dy,z+dz)+Newell_f(0,y-dy,z+dz)
	       +Newell_f(0,y+dy,z-dz)+Newell_f(0,y-dy,z-dz)));
	rycomp[index] += scale * 
	  (4*Newell_f(y,dx,z)
	   -2*(Newell_f(y+dy,dx,z)+Newell_f(y-dy,dx,z)
	       +Newell_f(y,dx,z+dz)+Newell_f(y,dx,z-dz))
	   +(Newell_f(y+dy,dx,z+dz)+Newell_f(y-dy,dx,z+dz)
	     +Newell_f(y+dy,dx,z-dz)+Newell_f(y-dy,dx,z-dz))
	   -8*Newell_f(y,0,z)
	   +4*(Newell_f(y+dy,0,z)+Newell_f(y-dy,0,z)
	       +Newell_f(y,0,z+dz)+Newell_f(y,0,z-dz))
	   -2*(Newell_f(y+dy,0,z+dz)+Newell_f(y-dy,0,z+dz)
	       +Newell_f(y+dy,0,z-dz)+Newell_f(y-dy,0,z-dz)));
	rzcomp[index] += scale * 
	  (4*Newell_f(z,y,dx)
	   -2*(Newell_f(z,y+dy,dx)+Newell_f(z,y-dy,dx)
	       +Newell_f(z+dz,y,dx)+Newell_f(z-dz,y,dx))
	   +(Newell_f(z+dz,y+dy,dx)+Newell_f(z+dz,y-dy,dx)
	     +Newell_f(z-dz,y+dy,dx)+Newell_f(z-dz,y-dy,dx))
	   -8*Newell_f(z,y,0)
	   +4*(Newell_f(z,y+dy,0)+Newell_f(z,y-dy,0)
	       +Newell_f(z+dz,y,0)+Newell_f(z-dz,y,0))
	   -2*(Newell_f(z+dz,y+dy,0)+Newell_f(z+dz,y-dy,0)
	       +Newell_f(z-dz,y+dy,0)+Newell_f(z-dz,y-dy,0)));
      }
    }
  } else {
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      for(j=0;j<rdimy;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<rdimx;i++) {
	  index = i+jkindex;
	  rxcomp[index] +=
	    -2*rxcomp[index+1] + rxcomp[index+2];
	  rycomp[index] +=
	    -2*rycomp[index+1] + rycomp[index+2];
	  rzcomp[index] +=
	    -2*rzcomp[index+1] + rzcomp[index+2];
	}
      }
    }
    for(k=0;k<rdimz;k++) { // Zero-fill overhang
      UINT4m kindex = k*rstridez;
      for(j=0;j<rdimy;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=rdimx;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] = 0.0;
	  rycomp[index] = 0.0;
	  rzcomp[index] = 0.0;
	}
      }
    }
  }

  // Step 3: Use symmetries to reflect into other octants.
  for(k=0;k<rdimz;k++) {
    UINT4m kindex = k*rstridez;
    for(j=0;j<rdimy;j++) {
      UINT4m jkindex = kindex + j*rstridey;
      for(i=0;i<rdimx;i++) {
	index = i+jkindex;
	REALWIDE xtemp = rxcomp[index];
	REALWIDE ytemp = rycomp[index];
	REALWIDE ztemp = rzcomp[index];
	if(i>0) {
	  UINT4m tindex = (2*cdimx-i)+j*rstridey+k*rstridez;
	  rxcomp[tindex]=xtemp;
	  rycomp[tindex]=ytemp;
	  rzcomp[tindex]=ztemp;
	}
	if(j>0) {
	  UINT4m tindex = i+(cdimy-j)*rstridey+k*rstridez;
	  rxcomp[tindex]=xtemp;
	  rycomp[tindex]=ytemp;
	  rzcomp[tindex]=ztemp;
	}
	if(k>0) {
	  UINT4m tindex = i+j*rstridey+(cdimz-k)*rstridez;
	  rxcomp[tindex]=xtemp;
	  rycomp[tindex]=ytemp;
	  rzcomp[tindex]=ztemp;
	}
	if(i>0 && j>0) {
	  UINT4m tindex = (2*cdimx-i)+(cdimy-j)*rstridey+k*rstridez;
	  rxcomp[tindex]=xtemp;
	  rycomp[tindex]=ytemp;
	  rzcomp[tindex]=ztemp;
	}
	if(i>0 && k>0) {
	  UINT4m tindex = (2*cdimx-i)+j*rstridey+(cdimz-k)*rstridez;
	  rxcomp[tindex]=xtemp;
	  rycomp[tindex]=ytemp;
	  rzcomp[tindex]=ztemp;
	}
	if(j>0 && k>0) {
	  UINT4m tindex = i+(cdimy-j)*rstridey+(cdimz-k)*rstridez;
	  rxcomp[tindex]=xtemp;
	  rycomp[tindex]=ytemp;
	  rzcomp[tindex]=ztemp;
	}
	if(i>0 && j>0 && k>0) {
	  UINT4m tindex = (2*cdimx-i)+(cdimy-j)*rstridey+(cdimz-k)*rstridez;
	  rxcomp[tindex]=xtemp;
	  rycomp[tindex]=ytemp;
	  rzcomp[tindex]=ztemp;
	}
      }
    }
  }

  // Special "SelfDemag" code may be more accurate at index 0,0,0
  rxcomp[0] = SelfDemagNx(dx,dy,dz)/(-2.0*cdimx*cdimy*cdimz);
  rycomp[0] = SelfDemagNy(dx,dy,dz)/(-2.0*cdimx*cdimy*cdimz);
  rzcomp[0] = SelfDemagNz(dx,dy,dz)/(-2.0*cdimx*cdimy*cdimz);

  // Step 4: Transform into frequency domain.
  fft.ForwardRealDataNoScale(xcomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(ycomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(zcomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);

  // Copy results from ?comp into A??.  We only need store 1/8th
  // of the results because of symmetries.
  for(k=0;k<adimz;k++) for(j=0;j<adimy;j++) for(i=0;i<adimx-1;i++) {
    UINT4m aindex = i+j*astridey+k*astridez;
    index = i+j*cstridey+k*cstridez;
    A00[aindex] = xcomp[index].real();
    A11[aindex] = ycomp[index].real();
    A22[aindex] = zcomp[index].real();
  }

  // Handle special packing for i=adimx-1 plane
  for(k=0;k<adimz;k++) for(j=0;j<adimy;j++) {
    Oxs_Complex temp00 = fft.RetrievePackedIndex(xcomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    Oxs_Complex temp11 = fft.RetrievePackedIndex(ycomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    Oxs_Complex temp22 = fft.RetrievePackedIndex(zcomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    UINT4m aindex = (adimx-1)+j*astridey+k*astridez;
    A00[aindex] = temp00.real();
    A11[aindex] = temp11.real();
    A22[aindex] = temp22.real();
  }

  // Repeat for Nxy, Nxz and Nyz. //////////////////////////////////////
  // Step 1: Evaluate g at each cell site.  Offset by (-dx,-dy,-dz)
  //  so we can do 2nd derivative operations "in-place".
  for(index=0;index<ctotalsize;index++) {
    // Reset temp space to zeros
    xcomp[index].Set(0.,0.);
    ycomp[index].Set(0.,0.);
    zcomp[index].Set(0.,0.);
  }
  for(k=0;k<kstop;k++) {
    UINT4m kindex = k*rstridez;
    REALWIDE z = 0.0;
    if(k==0) z = -dz;
    else if(k>1) z = dz*(k-1);
    for(j=0;j<jstop;j++) {
      UINT4m jkindex = kindex + j*rstridey;
      REALWIDE y = 0.0;
      if(j==0) y = -dy;
      else if(j>1) y = dy*(j-1);
      for(i=0;i<istop;i++) {
	index = i+jkindex;
	REALWIDE x = 0.0;
	if(i==0) x = -dx;
	else if(i>1) x = dx*(i-1);
	rxcomp[index] = scale*Newell_g(x,y,z);
	rycomp[index] = scale*Newell_g(x,z,y);
	rzcomp[index] = scale*Newell_g(y,z,x);
      }
    }
  }

  // Step 2a: Do d^2/dz^2
  if(kstop==1) {
    // Only 1 layer in z-direction of f stored in r?comp arrays.
    for(j=0;j<jstop;j++) {
      UINT4m jkindex = j*rstridey;
      REALWIDE y = 0.0;
      if(j==0) y = -dy;
      else if(j>1) y = dy*(j-1);
      for(i=0;i<istop;i++) {
	index = i+jkindex;
	REALWIDE x = 0.0;
	if(i==0) x = -dx;
	else if(i>1) x = dx*(i-1);
	rxcomp[index] += scale*(Newell_g(x,y,dz)-2*Newell_g(x,y,0));
	// If zdim==1, then rycomp and rzcomp, i.e., Nxz and Nyz,
	// are identically zero.  Don't bother setting the array
	// values here, because they get set to 0 in Step 3 below.
      }
    }
  } else {
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      for(j=0;j<jstop;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] +=
	    -2*rxcomp[index+rstridez] + rxcomp[index+2*rstridez];
	  rycomp[index] +=
	    -2*rycomp[index+rstridez] + rycomp[index+2*rstridez];
	  rzcomp[index] +=
	    -2*rzcomp[index+rstridez] + rzcomp[index+2*rstridez];
	}
      }
    }
    for(k=rdimz;k<kstop;k++) { // Zero-fill overhang
      UINT4m kindex = k*rstridez;
      for(j=0;j<jstop;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] = 0.0;
	  rycomp[index] = 0.0;
	  rzcomp[index] = 0.0;
	}
      }
    }
  }
  // Step 2b: Do d^2/dy^2
  if(jstop==1) {
    // Only 1 layer in y-direction of f stored in r?comp arrays.
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      REALWIDE z = dz*k;
      for(i=0;i<istop;i++) {
	index = i+kindex;
	REALWIDE x = 0.0;
	if(i==0) x = -dx;
	else if(i>1) x = dx*(i-1);
	rycomp[index] += scale * 
	  ((Newell_g(x,z-dz,dy)-2*Newell_g(x,z,dy)+Newell_g(x,z+dz,dy))
	   -2*(Newell_g(x,z-dz,0)-2*Newell_g(x,z,0)+Newell_g(x,z+dz,0)));
	// If ydim==1, then rxcomp and rzcomp, i.e., Nxy and Nyz,
	// are identically zero.  Don't bother setting the array
	// values here, because they get set to 0 in Step 3 below.
      }
    }
  } else {
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      for(j=0;j<rdimy;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] +=
	    -2*rxcomp[index+rstridey] + rxcomp[index+2*rstridey];
	  rycomp[index] +=
	    -2*rycomp[index+rstridey] + rycomp[index+2*rstridey];
	  rzcomp[index] +=
	    -2*rzcomp[index+rstridey] + rzcomp[index+2*rstridey];
	}
      }
    }
    for(k=0;k<rdimz;k++) { // Zero-fill overhang
      UINT4m kindex = k*rstridez;
      for(j=rdimy;j<jstop;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] = 0.0;
	  rycomp[index] = 0.0;
	  rzcomp[index] = 0.0;
	}
      }
    }
  }
  // Step 2c: Do d^2/dx^2
  if(istop==1) {
    // Only 1 layer in x-direction of f stored in r?comp arrays.
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      REALWIDE z = dz*k;
      for(j=0;j<rdimy;j++) {
	index = kindex + j*rstridey;
	REALWIDE y = dy*j;
	rzcomp[index] += scale * 
	  (4*Newell_g(y,z,dx)
	   -2*(Newell_g(y+dy,z,dx)+Newell_g(y-dy,z,dx)
	       +Newell_g(y,z+dz,dx)+Newell_g(y,z-dz,dx))
	   +(Newell_g(y+dy,z+dz,dx)+Newell_g(y-dy,z+dz,dx)
	     +Newell_g(y+dy,z-dz,dx)+Newell_g(y-dy,z-dz,dx))
	   -8*Newell_g(y,z,0)
	   +4*(Newell_g(y+dy,z,0)+Newell_g(y-dy,z,0)
	       +Newell_g(y,z+dz,0)+Newell_g(y,z-dz,0))
	   -2*(Newell_g(y+dy,z+dz,0)+Newell_g(y-dy,z+dz,0)
	       +Newell_g(y+dy,z-dz,0)+Newell_g(y-dy,z-dz,0)));
	// If xdim==1, then rxcomp and rycomp, i.e., Nxy and Nxz,
	// are identically zero.  Don't bother setting the array
	// values here, because they get set to 0 in Step 3 below.
      }
    }
  } else {
    for(k=0;k<rdimz;k++) {
      UINT4m kindex = k*rstridez;
      for(j=0;j<rdimy;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=0;i<rdimx;i++) {
	  index = i+jkindex;
	  rxcomp[index] +=
	    -2*rxcomp[index+1] + rxcomp[index+2];
	  rycomp[index] +=
	    -2*rycomp[index+1] + rycomp[index+2];
	  rzcomp[index] +=
	    -2*rzcomp[index+1] + rzcomp[index+2];
	}
      }
    }
    for(k=0;k<rdimz;k++) { // Zero-fill overhang
      UINT4m kindex = k*rstridez;
      for(j=0;j<rdimy;j++) {
	UINT4m jkindex = kindex + j*rstridey;
	for(i=rdimx;i<istop;i++) {
	  index = i+jkindex;
	  rxcomp[index] = 0.0;
	  rycomp[index] = 0.0;
	  rzcomp[index] = 0.0;
	}
      }
    }
  }

  // Step 3: Use symmetries to reflect into other octants.
  //     Also, at each coordinate plane, set to 0.0 any term
  //     which is odd across that boundary.  It should already
  //     be close to 0, but will likely be slightly off due to
  //     rounding errors.
  for(k=0;k<rdimz;k++) {
    UINT4m kindex = k*rstridez;
    for(j=0;j<rdimy;j++) {
      UINT4m jkindex = kindex + j*rstridey;
      for(i=0;i<rdimx;i++) {
	index = i+jkindex;

	if(i==0 || j==0) rxcomp[index]=0.0;
	if(i==0 || k==0) rycomp[index]=0.0;
	if(j==0 || k==0) rzcomp[index]=0.0;

	REALWIDE xtemp = rxcomp[index];
	REALWIDE ytemp = rycomp[index];
	REALWIDE ztemp = rzcomp[index];
	if(i>0) {
	  UINT4m tindex = (2*cdimx-i)+j*rstridey+k*rstridez;
	  rxcomp[tindex] = -1*xtemp;
	  rycomp[tindex] = -1*ytemp;
	  rzcomp[tindex] =    ztemp;
	}
	if(j>0) {
	  UINT4m tindex = i+(cdimy-j)*rstridey+k*rstridez;
	  rxcomp[tindex] = -1*xtemp;
	  rycomp[tindex] =    ytemp;
	  rzcomp[tindex] = -1*ztemp;
	}
	if(k>0) {
	  UINT4m tindex = i+j*rstridey+(cdimz-k)*rstridez;
	  rxcomp[tindex] =    xtemp;
	  rycomp[tindex] = -1*ytemp;
	  rzcomp[tindex] = -1*ztemp;
	}
	if(i>0 && j>0) {
	  UINT4m tindex = (2*cdimx-i)+(cdimy-j)*rstridey+k*rstridez;
	  rxcomp[tindex] =    xtemp;
	  rycomp[tindex] = -1*ytemp;
	  rzcomp[tindex] = -1*ztemp;
	}
	if(i>0 && k>0) {
	  UINT4m tindex = (2*cdimx-i)+j*rstridey+(cdimz-k)*rstridez;
	  rxcomp[tindex] = -1*xtemp;
	  rycomp[tindex] =    ytemp;
	  rzcomp[tindex] = -1*ztemp;
	}
	if(j>0 && k>0) {
	  UINT4m tindex = i+(cdimy-j)*rstridey+(cdimz-k)*rstridez;
	  rxcomp[tindex] = -1*xtemp;
	  rycomp[tindex] = -1*ytemp;
	  rzcomp[tindex] =    ztemp;
	}
	if(i>0 && j>0 && k>0) {
	  UINT4m tindex = (2*cdimx-i)+(cdimy-j)*rstridey+(cdimz-k)*rstridez;
	  rxcomp[tindex] = xtemp;
	  rycomp[tindex] = ytemp;
	  rzcomp[tindex] = ztemp;
	}
      }
    }
  }

  // Step 4: Transform into frequency domain.
  fft.ForwardRealDataNoScale(xcomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(ycomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(zcomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);

  // Copy results from ?comp into A??.  We only need store 1/8th
  // of the results because of symmetries.
  for(k=0;k<adimz;k++) for(j=0;j<adimy;j++) for(i=0;i<adimx-1;i++) {
    UINT4m aindex = i+j*astridey+k*astridez;
    index = i+j*cstridey+k*cstridez;
    A01[aindex] = xcomp[index].real();
    A02[aindex] = ycomp[index].real();
    A12[aindex] = zcomp[index].real();
  }
  // Handle special packing for i=adimx-1 plane
  for(k=0;k<adimz;k++) for(j=0;j<adimy;j++) {
    Oxs_Complex temp01 = fft.RetrievePackedIndex(xcomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    Oxs_Complex temp02 = fft.RetrievePackedIndex(ycomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    Oxs_Complex temp12 = fft.RetrievePackedIndex(zcomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    UINT4m aindex = (adimx-1)+j*astridey+k*astridez;
    A01[aindex] = temp01.real();
    A02[aindex] = temp02.real();
    A12[aindex] = temp12.real();
  }
}

void Oxs_Demag::FillCoefficientArrays_OLD(const Oxs_Mesh* genmesh) const
{ // This routine is conceptually const.
  // This "OLD" variant is simpler but slower, and is retained
  // solely for testing and development purposes.

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(genmesh);
  if(mesh==NULL) {
    string msg=string("Object ")
      + string(genmesh->InstanceName())
      + string(" is not a rectangular mesh.");
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Fill dimension variables
  rdimx = mesh->DimX();
  rdimy = mesh->DimY();
  rdimz = mesh->DimZ();
  if(rdimx==0 || rdimy==0 || rdimz==0) return; // Empty mesh!

  cdimx = NextPowerOfTwo(2*rdimx)/2;
  if(rdimy>1) cdimy = NextPowerOfTwo(2*rdimy);
  else        cdimy = 1;
  if(rdimz>1) cdimz = NextPowerOfTwo(2*rdimz);
  else        cdimz = 1;
  cstridey = cdimx + 1;  // Pad by one to avoid cache line entanglement
  cstridez = cstridey*cdimy;

  UINT4m ctotalsize=cstridez*cdimz;
  UINT4m rtotalsize=2*ctotalsize;
  if(rtotalsize<2*cdimx || rtotalsize<cdimy || rtotalsize<cdimz) {
    // Partial overflow check
    string msg="UINT4m overflow in " + string(InstanceName())
      + ": Product cdimx*cdimy*cdimz too big to fit in a UINT4m variable";
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Allocate memory for interaction matrices and magnetization components
  adimx=1+cdimx;
  adimy=1+cdimy/2;
  adimz=1+cdimz/2;
  astridey=adimx;
  astridez=adimy*astridey;
  UINT4m atotalsize=adimz*astridez;
  A00   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A01   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A02   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A11   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A12   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  A22   = new OXS_COMPLEX_REAL_TYPE[atotalsize];
  xcomp = new Oxs_Complex[ctotalsize];
  ycomp = new Oxs_Complex[ctotalsize];
  zcomp = new Oxs_Complex[ctotalsize];
  if(xcomp==NULL || ycomp==NULL || zcomp==NULL
     || A00==NULL || A01==NULL || A02==NULL
     || A11==NULL || A12==NULL || A22==NULL) {
    // Safety check for those machines on which new[] doesn't throw
    // BadAlloc.
    string msg="Insufficient memory in Demag constructor.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  REALWIDE *rxcomp = static_cast<REALWIDE*>(static_cast<void*>(xcomp));
  REALWIDE *rycomp = static_cast<REALWIDE*>(static_cast<void*>(ycomp));
  REALWIDE *rzcomp = static_cast<REALWIDE*>(static_cast<void*>(zcomp));
  UINT4m rstridey=2*cstridey;
  UINT4m rstridez=2*cstridez;

  // According (16) in Newell's paper, the demag field is given by
  //                        H = -N*M
  // where N is the "demagnetizing tensor," with components Nxx, Nxy,
  // etc.  With the '-1' in 'scale' we store '-N' instead of 'N',
  // so we don't have to multiply the output from the FFT + iFFT
  // by -1 in ConstMagField() below.

  // Fill interaction matrices with demag coefs from Newell's paper.
  // Note that A00, A11 and A22 are even in x,y and z.
  // A01 is odd in x and y, even in z.
  // A02 is odd in x and z, even in y.
  // A12 is odd in y and z, even in x.
  // We use these symmetries to reduce storage requirements.  If
  // f is real and even, then f^ is also real and even.  If f
  // is real and odd, then f^ is (pure) imaginary and odd.
  // As a result, the transform of each of the Axx interaction
  // matrices will be real, with the same even/odd properties.

  UINT4m index,i,j,k;

  REALWIDE dx = mesh->EdgeLengthX();
  REALWIDE dy = mesh->EdgeLengthY();
  REALWIDE dz = mesh->EdgeLengthZ();
  // For demag calculation, all that matters is the relative
  // sizes of dx, dy and dz.  To help insure we don't run
  // outside floating point range, rescale these values so
  // largest is 1.0
  REALWIDE maxedge=dx;
  if(dy>maxedge) maxedge=dy;
  if(dz>maxedge) maxedge=dz;
  dx/=maxedge; dy/=maxedge; dz/=maxedge;

  REALWIDE scale = -1./(4*PI*dx*dy*dz);

  // Also throw in FFT scaling.  This allows the "NoScale" FFT routines
  // to be used.
  scale *= 1.0/(2*cdimx*cdimy*cdimz);

  for(index=0;index<ctotalsize;index++) {
    // Initialize temp space to 0
    xcomp[index].Set(0.,0.);
    ycomp[index].Set(0.,0.);
    zcomp[index].Set(0.,0.);
  }


#ifdef DUMP_COEF_TEST
fprintf(stderr,"Nxy(1,2,3,1,2,3)=%.17g   Nxy(10,1,1,1,2,3)=%.17g\n",
	(double)CalculateSDA01(1.,2.,3.,1.,2.,3.)/(4*PI*1.*2.*3.),
	(double)CalculateSDA01(10.,1.,1.,1.,2.,3.)/(4*PI*1.*2.*3.));
fprintf(stderr,"Nxy(-1,2,3,1,2,3)=%.17g   Nxy(10,1,-1,1,2,3)=%.17g\n",
	(double)CalculateSDA01(-1.,2.,3.,1.,2.,3.)/(4*PI*1.*2.*3.),
	(double)CalculateSDA01(10.,1.,-1.,1.,2.,3.)/(4*PI*1.*2.*3.));
fprintf(stderr,"Nxy(1,1,0,1,2,3)=%.17g   Nxy(1,1,0,2,1,3)=%.17g\n",
	(double)CalculateSDA01(1.,1.,0.,1.,2.,3.)/(4*PI*1.*2.*3.),
	(double)CalculateSDA01(1.,1.,0.,2.,1.,3.)/(4*PI*1.*2.*3.));
fprintf(stderr,"Nxy(-1,1,0,1,2,3)=%.17g   Nxy(1,-1,0,2,1,3)=%.17g\n",
	(double)CalculateSDA01(-1.,1.,0.,1.,2.,3.)/(4*PI*1.*2.*3.),
	(double)CalculateSDA01(1.,-1.,0.,2.,1.,3.)/(4*PI*1.*2.*3.));
#endif // DUMP_COEF_TEST

  for(k=0;k<rdimz;k++) for(j=0;j<rdimy;j++) for(i=0;i<rdimx;i++) {
    REALWIDE x = dx*i;
    REALWIDE y = dy*j;
    REALWIDE z = dz*k;
    REALWIDE a00=scale*CalculateSDA00(x,y,z,dx,dy,dz);
    REALWIDE a01=scale*CalculateSDA01(x,y,z,dx,dy,dz);
    REALWIDE a02=scale*CalculateSDA02(x,y,z,dx,dy,dz);

    index = i+j*rstridey+k*rstridez;

    rxcomp[index]=a00;
    rycomp[index]=a01;
    rzcomp[index]=a02;

    if(i>0) {
      UINT4m tindex = (2*cdimx-i)+j*rstridey+k*rstridez;
      rxcomp[tindex]=a00;
      rycomp[tindex]=-a01;
      rzcomp[tindex]=-a02;
    }
    if(j>0) {
      UINT4m tindex = i+(cdimy-j)*rstridey+k*rstridez;
      rxcomp[tindex]=a00;
      rycomp[tindex]=-a01;
      rzcomp[tindex]=a02;
    }
    if(k>0) {
      UINT4m tindex = i+j*rstridey+(cdimz-k)*rstridez;
      rxcomp[tindex]=a00;
      rycomp[tindex]=a01;
      rzcomp[tindex]=-a02;
    }
    if(i>0 && j>0) {
      UINT4m tindex = (2*cdimx-i)+(cdimy-j)*rstridey+k*rstridez;
      rxcomp[tindex]=a00;
      rycomp[tindex]=a01;
      rzcomp[tindex]=-a02;
    }
    if(i>0 && k>0) {
      UINT4m tindex = (2*cdimx-i)+j*rstridey+(cdimz-k)*rstridez;
      rxcomp[tindex]=a00;
      rycomp[tindex]=-a01;
      rzcomp[tindex]=a02;
    }
    if(j>0 && k>0) {
      UINT4m tindex = i+(cdimy-j)*rstridey+(cdimz-k)*rstridez;
      rxcomp[tindex]=a00;
      rycomp[tindex]=-a01;
      rzcomp[tindex]=-a02;
    }
    if(i>0 && j>0 && k>0) {
      UINT4m tindex = (2*cdimx-i)+(cdimy-j)*rstridey+(cdimz-k)*rstridez;
      rxcomp[tindex]=a00;
      rycomp[tindex]=a01;
      rzcomp[tindex]=a02;
    }
  }

  // Transform into frequency domain.
  fft.ForwardRealDataNoScale(xcomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(ycomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(zcomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);

  // Copy results from ?comp into A??.  We only need store 1/8th
  // of the results because of symmetries.
  for(k=0;k<adimz;k++) for(j=0;j<adimy;j++) for(i=0;i<adimx-1;i++) {
    UINT4m aindex = i+j*astridey+k*astridez;
    index = i+j*cstridey+k*cstridez;
    A00[aindex] = xcomp[index].real();
    A01[aindex] = ycomp[index].real();
    A02[aindex] = zcomp[index].real();
  }
  // Handle special packing for i=adimx-1 plane
  for(k=0;k<adimz;k++) for(j=0;j<adimy;j++) {
    Oxs_Complex temp00 = fft.RetrievePackedIndex(xcomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    Oxs_Complex temp01 = fft.RetrievePackedIndex(ycomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    Oxs_Complex temp02 = fft.RetrievePackedIndex(zcomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    UINT4m aindex = (adimx-1)+j*astridey+k*astridez;
    A00[aindex] = temp00.real();
    A01[aindex] = temp01.real();
    A02[aindex] = temp02.real();
  }

  // Repeat for A11, A12 and A22. //////////////////////////////////////
  for(index=0;index<ctotalsize;index++) {
    xcomp[index].Set(0.,0.);
    ycomp[index].Set(0.,0.);
    zcomp[index].Set(0.,0.);
  }

  for(k=0;k<rdimz;k++) for(j=0;j<rdimy;j++) for(i=0;i<rdimx;i++) {
    REALWIDE x = dx*i;    REALWIDE y = dy*j;    REALWIDE z = dz*k;
    REALWIDE a11=scale*CalculateSDA11(x,y,z,dx,dy,dz);
    REALWIDE a12=scale*CalculateSDA12(x,y,z,dx,dy,dz);
    REALWIDE a22=scale*CalculateSDA22(x,y,z,dx,dy,dz);
    index = i+j*rstridey+k*rstridez;
    rxcomp[index]=a11;
    rycomp[index]=a12;
    rzcomp[index]=a22;
    if(i>0) {
      UINT4m tindex = (2*cdimx-i)+j*rstridey+k*rstridez;
      rxcomp[tindex]=a11;
      rycomp[tindex]=a12;
      rzcomp[tindex]=a22;
    }
    if(j>0) {
      UINT4m tindex = i+(cdimy-j)*rstridey+k*rstridez;
      rxcomp[tindex]=a11;
      rycomp[tindex]=-a12;
      rzcomp[tindex]=a22;
    }
    if(k>0) {
      UINT4m tindex = i+j*rstridey+(cdimz-k)*rstridez;
      rxcomp[tindex]=a11;
      rycomp[tindex]=-a12;
      rzcomp[tindex]=a22;
    }
    if(i>0 && j>0) {
      UINT4m tindex = (2*cdimx-i)+(cdimy-j)*rstridey+k*rstridez;
      rxcomp[tindex]=a11;
      rycomp[tindex]=-a12;
      rzcomp[tindex]=a22;
    }
    if(i>0 && k>0) {
      UINT4m tindex = (2*cdimx-i)+j*rstridey+(cdimz-k)*rstridez;
      rxcomp[tindex]=a11;
      rycomp[tindex]=-a12;
      rzcomp[tindex]=a22;
    }
    if(j>0 && k>0) {
      UINT4m tindex = i+(cdimy-j)*rstridey+(cdimz-k)*rstridez;
      rxcomp[tindex]=a11;
      rycomp[tindex]=a12;
      rzcomp[tindex]=a22;
    }
    if(i>0 && j>0 && k>0) {
      UINT4m tindex = (2*cdimx-i)+(cdimy-j)*rstridey+(cdimz-k)*rstridez;
      rxcomp[tindex]=a11;
      rycomp[tindex]=a12;
      rzcomp[tindex]=a22;
    }
  }

  fft.ForwardRealDataNoScale(xcomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(ycomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(zcomp,2*cdimx,cdimy,cdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);

  for(k=0;k<adimz;k++) for(j=0;j<adimy;j++) for(i=0;i<adimx-1;i++) {
    UINT4m aindex = i+j*astridey+k*astridez;
    index = i+j*cstridey+k*cstridez;
    A11[aindex] = xcomp[index].real();
    A12[aindex] = ycomp[index].real();
    A22[aindex] = zcomp[index].real();
  }
  for(k=0;k<adimz;k++) for(j=0;j<adimy;j++) { // Special packing case
    Oxs_Complex temp11 = fft.RetrievePackedIndex(xcomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    Oxs_Complex temp12 = fft.RetrievePackedIndex(ycomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    Oxs_Complex temp22 = fft.RetrievePackedIndex(zcomp,
		   cdimx,cdimy,cdimz,cstridey,cstridez,
		   adimx-1,j,k);
    UINT4m aindex = (adimx-1)+j*astridey+k*astridez;
    A11[aindex] = temp11.real();
    A12[aindex] = temp12.real();
    A22[aindex] = temp22.real();
  }
}

void Oxs_Demag::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  // (Re)-initialize mesh coefficient array if mesh has changed.
  if(mesh_id != state.mesh->Id()) {
    mesh_id = 0; // Safety
    if(kernel_fill_routine.compare("old")==0) {
      FillCoefficientArrays_OLD(state.mesh);
    } else {
      FillCoefficientArrays(state.mesh);
    }
    mesh_id = state.mesh->Id();
  }

  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  // Calculate FFT of Mx, My and Mz
  UINT4m i,j,k;
  UINT4m mstridey=rdimx;  // Assume import mesh is tight packed
  UINT4m mstridez=rdimy*mstridey;
  UINT4m rstridey=2*cstridey;
  UINT4m rstridez=2*cstridez;
  OXS_COMPLEX_REAL_TYPE* rxcomp
    = static_cast<OXS_COMPLEX_REAL_TYPE*>(static_cast<void*>(xcomp));
  OXS_COMPLEX_REAL_TYPE* rycomp
    = static_cast<OXS_COMPLEX_REAL_TYPE*>(static_cast<void*>(ycomp));
  OXS_COMPLEX_REAL_TYPE* rzcomp
    = static_cast<OXS_COMPLEX_REAL_TYPE*>(static_cast<void*>(zcomp));

  for(k=0;k<rdimz;k++) for(j=0;j<rdimy;j++) {
    UINT4m mindex = j*mstridey+k*mstridez;
    UINT4m rindex = j*rstridey+k*rstridez;
    for(i=0;i<rdimx;i++) {
      REAL8m scale = Ms[mindex];
      const ThreeVector& vec = spin[mindex];
      ++mindex;
      rxcomp[rindex]=scale*vec.x;
      rycomp[rindex]=scale*vec.y;
      rzcomp[rindex]=scale*vec.z;
      ++rindex;
    }
  }

  fft.ForwardRealDataNoScale(xcomp,rdimx,rdimy,rdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(ycomp,rdimx,rdimy,rdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.ForwardRealDataNoScale(zcomp,rdimx,rdimy,rdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);

  // Calculate field components in frequency domain, then to iFFT to
  // transform back to space domain.  Make use of realness and even/odd
  // properties of interaction matrices Axx.
  UINT4m block;
  for(block=0;block<4;block++) {
    UINT4m base_offset = 0;
    if(block%2==1) {
      if(cdimy==1) continue; // There is no second "j" block
      base_offset += cstridey*(cdimy/2);
    }
    if(block/2==1) {
      if(cdimz==1) continue; // There is no second "k" block
      base_offset += cstridez*(cdimz/2);
    }
    Oxs_Complex* xbase=xcomp+base_offset;
    Oxs_Complex* ybase=ycomp+base_offset;
    Oxs_Complex* zbase=zcomp+base_offset;
    int jsign=1; if(block%2==1) jsign=-1;
    int ksign=1; if(block/2==1) ksign=-1;
    Oxs_Complex Mx,My,Mz;
    OXS_COMPLEX_REAL_TYPE a00,a11,a22,a01,a02,a12;
    UINT4m aindex=0,cindex;
    k=0;
    do {
      // j==0, i==0
      cindex = k*cstridez;
      Mx = xbase[cindex];
      My = ybase[cindex];
      Mz = zbase[cindex];
      if(k==0) {
	aindex = 0;
	if(jsign != 1) aindex += (adimy-1)*astridey;
	if(ksign != 1) aindex += (adimz-1)*astridez;
	OXS_COMPLEX_REAL_TYPE a00b,a11b,a22b,a01b,a02b,a12b;
	a00  = A00[aindex];
	a00b = A00[aindex+cdimx];
	a11  = A11[aindex];
	a11b = A11[aindex+cdimx];
	a22  = A22[aindex];
	a22b = A22[aindex+cdimx];
	a01  = A01[aindex];
	a01b = A01[aindex+cdimx];
	a02  = A02[aindex];
	a02b = A02[aindex+cdimx];
	a12  = A12[aindex];
	a12b = A12[aindex+cdimx];
	xbase[cindex].re = a00 *Mx.re + a01 *My.re + a02 *Mz.re;  // Hx
	xbase[cindex].im = a00b*Mx.im + a01b*My.im + a02b*Mz.im;
	ybase[cindex].re = a01 *Mx.re + a11 *My.re + a12 *Mz.re;  // Hy
	ybase[cindex].im = a01b*Mx.im + a11b*My.im + a12b*Mz.im;
	zbase[cindex].re = a02 *Mx.re + a12 *My.re + a22 *Mz.re;  // Hz
	zbase[cindex].im = a02b*Mx.im + a12b*My.im + a22b*Mz.im;
      } else {
	// k>0
	if(ksign == 1) {
	  aindex = k*astridez;
	} else {
	  aindex = (adimz-1-k)*astridez+cdimx;
	}
	if(jsign != 1) aindex += (adimy-1)*astridey;
	a00 = A00[aindex];
	a11 = A11[aindex];
	a22 = A22[aindex];
	a01 = A01[aindex];
	a02 = ksign*A02[aindex];
	a12 = ksign*A12[aindex];
	xbase[cindex] = a00*Mx + a01*My + a02*Mz;  // Hx
	ybase[cindex] = a01*Mx + a11*My + a12*Mz;  // Hy
	zbase[cindex] = a02*Mx + a12*My + a22*Mz;  // Hz
      }
      // j==0, i>0
      if(ksign == 1) aindex = k*astridez;
      else           aindex = (adimz-1-k)*astridez;
      if(jsign != 1) aindex += (adimy-1)*astridey;
      for(i=1;i<cdimx;i++) {
	++cindex;
	Mx = xbase[cindex];
	My = ybase[cindex];
	Mz = zbase[cindex];
	++aindex;
	a00 = A00[aindex];
	a11 = A11[aindex];
	a22 = A22[aindex];
	a01 = A01[aindex];
	a02 = ksign*A02[aindex];
	a12 = ksign*A12[aindex];
	xbase[cindex] = a00*Mx + a01*My + a02*Mz;  // Hx
	ybase[cindex] = a01*Mx + a11*My + a12*Mz;  // Hy
	zbase[cindex] = a02*Mx + a12*My + a22*Mz;  // Hz
      }

      for(j=1;j<cdimy/2;j++) {
	// i==0
	cindex = j*cstridey+k*cstridez;
	Mx = xbase[cindex];
	My = ybase[cindex];
	Mz = zbase[cindex];
	if(ksign == 1) aindex = k*astridez;
	else           aindex = (adimz-1-k)*astridez;
	if(jsign == 1) aindex += j*astridey;
	else           aindex += (adimy-1-j)*astridey + cdimx;
	a00 = A00[aindex];
	a11 = A11[aindex];
	a22 = A22[aindex];
	a01 = jsign*A01[aindex];
	a02 = ksign*A02[aindex];
	a12 = jsign*ksign*A12[aindex];
	xbase[cindex] = a00*Mx + a01*My + a02*Mz;  // Hx
	ybase[cindex] = a01*Mx + a11*My + a12*Mz;  // Hy
	zbase[cindex] = a02*Mx + a12*My + a22*Mz;  // Hz
	if(jsign != 1) aindex -= cdimx;
	for(i=1;i<cdimx;i++) {
	  ++cindex;
	  Mx = xbase[cindex];
	  My = ybase[cindex];
	  Mz = zbase[cindex];
	  ++aindex;
	  a00 = A00[aindex];
	  a11 = A11[aindex];
	  a22 = A22[aindex];
	  a01 = jsign*A01[aindex];
	  a02 = ksign*A02[aindex];
	  a12 = jsign*ksign*A12[aindex];
	  xbase[cindex] = a00*Mx + a01*My + a02*Mz;  // Hx
	  ybase[cindex] = a01*Mx + a11*My + a12*Mz;  // Hy
	  zbase[cindex] = a02*Mx + a12*My + a22*Mz;  // Hz
	}
      }
    } while(++k<cdimz/2);  // For each block, do one pass through
    /// the above loop if cdimz is 1 or 2.  If cdimz is 1 then
    /// ksign will always be 1, because the upper half "blocks"
    /// are skipped.
  }

  fft.InverseRealDataNoScale(xcomp,rdimx,rdimy,rdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.InverseRealDataNoScale(ycomp,rdimx,rdimy,rdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);
  fft.InverseRealDataNoScale(zcomp,rdimx,rdimy,rdimz,
			     cdimx,cdimy,cdimz,cstridey,cstridez);

  OXS_COMPLEX_REAL_TYPE mult =  -0.5 * MU0;
  for(k=0;k<rdimz;k++) for(j=0;j<rdimy;j++) {
    UINT4m mindex = j*mstridey+k*mstridez;
    UINT4m rindex = j*rstridey+k*rstridez;
    for(i=0;i<rdimx;i++) {
      field[mindex].x=rxcomp[rindex];
      field[mindex].y=rycomp[rindex];
      field[mindex].z=rzcomp[rindex];
      // Calculate also pointwise energy density: -0.5*MU0*<M,H>
      OXS_COMPLEX_REAL_TYPE dot = spin[mindex]*field[mindex];
      energy[mindex] = mult * Ms[mindex] * dot;
      ++mindex; ++rindex;
    }
  }
}
