/* FILE: mindriver.cc            -*-Mode: c++-*-
 *
 * Example minimization Oxs_Driver class.
 *
 */

#include "nb.h"
#include "mindriver.h"
#include "director.h"
#include "simstate.h"
#include "minevolver.h"
#include "key.h"
#include "mesh.h"
#include "scalarfield.h"
#include "vectorfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_MinDriver);

/* End includes */

// Constructor
Oxs_MinDriver::Oxs_MinDriver(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Driver(name,newdtr,argstr),
    max_mxHxm_output_obj_ptr(NULL), total_energy_output_obj_ptr(NULL)
{
  // Process arguments
  // Get evolver name specification
  OXS_GET_INIT_EXT_OBJECT("evolver",Oxs_MinEvolver,evolver_obj);
  evolver_key.Set(evolver_obj.GetPtr());
  // Dependency lock on Oxs_MinEvolver object is
  // held until *this is destroyed.

  GetGroupedRealListInitValue("stopping_mxHxm",stopping_mxHxm); // Required

  OXS_GET_INIT_EXT_OBJECT("mesh",Oxs_Mesh,mesh_obj);
  mesh_key.Set(mesh_obj.GetPtr());  // Sets a dep lock

  Oxs_OwnedPointer<Oxs_ScalarField> Msinit;
  OXS_GET_INIT_EXT_OBJECT("Ms",Oxs_ScalarField,Msinit);

  OXS_GET_INIT_EXT_OBJECT("m0",Oxs_VectorField,m0);

  VerifyAllInitArgsUsed();

  // Fill Ms and Ms_inverse array, and verify that Ms is non-negative.
  Msinit->FillMeshValue(mesh_obj.GetPtr(),Ms);
  Ms_inverse.AdjustSize(mesh_obj.GetPtr());
  for(UINT4m i=0;i<mesh_obj->Size();i++) {
    if(Ms[i]<0.0) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Negative Ms value (%g) detected at mesh index %d.",
		  Ms[i],i);
      throw Oxs_Ext::Error(this,buf);
    } else if(Ms[i]==0.0) {
      Ms_inverse[i]=0.0; // Special case handling
    } else {
      Ms_inverse[i]=1.0/Ms[i];
    }
  }

  // Reserve space for initial state (see GetInitialState() below)
  director->ReserveSimulationStateRequest(1);
}

Oxs_ConstKey<Oxs_SimState>
Oxs_MinDriver::GetInitialState() const
{
  Oxs_Key<Oxs_SimState> initial_state;
  director->GetNewSimulationState(initial_state);
  Oxs_SimState& istate = initial_state.GetWriteReference();
  istate.previous_state_id     = 0;
  istate.iteration_count       = 0;
  istate.stage_number          = 0;
  istate.stage_iteration_count = 0;
  istate.stage_start_time      = -1.;  // Dummy value
  istate.stage_elapsed_time    = -1.;  // Dummy value
  istate.last_timestep         = -1.;  // Dummy value

  istate.mesh = mesh_key.GetPtr();

  istate.Ms = &Ms;
  istate.Ms_inverse = &Ms_inverse;
  m0->FillMeshValue(istate.mesh,istate.spin);

  // Insure that all spins are unit vectors
  UINT4m size = istate.spin.Size();
  for(UINT4m i=0;i<size;i++) istate.spin[i].MakeUnit();

  const Oxs_SimState& fstate = initial_state.GetReadReference();
  /// Release write lock. The read lock will be automatically
  /// released when the key "initial_state" is destroyed.

  fstate.AddDerivedData("Last energy",0.);

  return initial_state;
}


BOOL Oxs_MinDriver::Init()
{ 
  Oxs_Driver::Init();  // Run init routine in parent.
  /// This will call Oxs_MinDriver::GetInitialState().

  // Get pointer to output object providing max mxHxm data
  const Oxs_MinEvolver* evolver = evolver_key.GetPtr();
  if(evolver==NULL) {
    throw Oxs_Ext::Error(this,"PROGRAMMING ERROR: No evolver found?");
  }

  string output_name = string(evolver->InstanceName());
  output_name += string(":Max mxHxm");
  max_mxHxm_output_obj_ptr
    =  director->FindOutputObject(output_name.c_str());
  if(max_mxHxm_output_obj_ptr==NULL) {
    string msg = string("Unable to identify unique \"")
      + output_name + string("\" object.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  output_name = string(evolver->InstanceName());
  output_name += string(":Total energy");
  total_energy_output_obj_ptr
    =  director->FindOutputObject(output_name.c_str());
  if(max_mxHxm_output_obj_ptr==NULL) {
    string msg = string("Unable to identify unique \"")
      + output_name + string("\" object.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  return 1;
}

Oxs_MinDriver::~Oxs_MinDriver()
{}


void Oxs_MinDriver::StageRequestCount
(unsigned int& min,
 unsigned int& max) const
{ // Number of stages wanted by driver

  Oxs_Driver::StageRequestCount(min,max);

  unsigned int count = stopping_mxHxm.size();
  if(count>min) min=count;
  if(count>1 && count<max) max=count;
  // Treat length 1 lists as imposing no upper constraint.
}

BOOL
Oxs_MinDriver::ChildIsStageDone(const Oxs_SimState& state) const
{
  UINT4m stage_index = state.stage_number;

  // Max mxHxm check
  Tcl_Interp* mif_interp = director->GetMifInterp();
  if(max_mxHxm_output_obj_ptr == NULL ||
     max_mxHxm_output_obj_ptr->Output(&state,mif_interp,0,NULL) != TCL_OK) {
    string msg=string("Unable to obtain Max mxHxm output: ");
    if(max_mxHxm_output_obj_ptr==NULL) {
      msg += string("PROGRAMMING ERROR: max_mxHxm_output_obj_ptr not set."
		    " Driver Init() probably not called.");
    } else {
      msg += string(Tcl_GetStringResult(mif_interp));
    }
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  BOOL err;
  REAL8m max_mxHxm = Nb_Atof(Tcl_GetStringResult(mif_interp),err);
  if(err) {
    string msg=string("Error detected in StageDone method"
		      " --- Invalid Max dm/dt output: ");
    msg += string(Tcl_GetStringResult(mif_interp));
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  REAL8m stop_mxHxm=0.;
  if(stage_index >= stopping_mxHxm.size()) {
    stop_mxHxm = stopping_mxHxm[stopping_mxHxm.size()-1];
  } else {
    stop_mxHxm = stopping_mxHxm[stage_index];
  }
  if(stop_mxHxm>0.0 && max_mxHxm <= stop_mxHxm) {
    return 1; // Stage done
  }

  // If control gets here, then stage not done
  return 0;
}

BOOL
Oxs_MinDriver::ChildIsRunDone(const Oxs_SimState& /* state */) const
{
  // No child-specific checks at this time...
  return 0; // Run not done
}

void Oxs_MinDriver::FillState
(const Oxs_SimState& old_state,
 Oxs_SimState& new_state) const
{
  Oxs_Driver::FillState(old_state,new_state);

  // Copy over values from old_state for time indexes, since
  // with mindriver time is not advanced.  Perhaps the time index
  // variables should be moved over to the WOO area in simstate?
  new_state.stage_start_time   = old_state.stage_start_time;
  new_state.stage_elapsed_time = old_state.stage_elapsed_time;
  new_state.last_timestep      = old_state.last_timestep;

  // Keep track of energy from previous state.  This is used
  // by "Delta E" output.
  REAL8m old_energy;
  if(!old_state.GetDerivedData("Total energy",old_energy)) {
    // Energy field not filled in old state.  Try calling output
    // object, which should call necessary update functions.
    Tcl_Interp* mif_interp = director->GetMifInterp();
    if(total_energy_output_obj_ptr == NULL ||
       total_energy_output_obj_ptr->Output(&old_state,mif_interp,0,NULL)
       != TCL_OK) {
      string msg =
	string("Error in FillState:"
	       " Unable to obtain \"Total energy\" output: ");
      if(total_energy_output_obj_ptr==NULL) {
	msg += string("PROGRAMMING ERROR:"
		      " total_energy_output_obj_ptr not set."
		      " Driver Init() probably not called.");
      } else {
	msg += string(Tcl_GetStringResult(mif_interp));
      }
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    BOOL err;
    old_energy = Nb_Atof(Tcl_GetStringResult(mif_interp),err);
    if(err) {
      string msg=string("Error detected in FillState method"
			" --- Invalid \"Total energy\" output: ");
      msg += string(Tcl_GetStringResult(mif_interp));
      throw Oxs_Ext::Error(this,msg.c_str());
    }
  }
  new_state.AddDerivedData("Last energy",old_energy);
}

void Oxs_MinDriver::FillNewStageState
(const Oxs_SimState& old_state,
 int new_stage_number,
 Oxs_SimState& new_state) const
{
  Oxs_Driver::FillNewStageState(old_state,new_stage_number,new_state);
  REAL8m old_energy;
  if(!old_state.GetDerivedData("Total energy",old_energy)) {
    // Energy field not filled in old state.  Try calling output
    // object, which should call necessary update functions.
    Tcl_Interp* mif_interp = director->GetMifInterp();
    if(total_energy_output_obj_ptr == NULL ||
       total_energy_output_obj_ptr->Output(&old_state,mif_interp,0,NULL)
       != TCL_OK) {
      string msg =
	string("Error in FillNewStageState:"
	       " Unable to obtain \"Total energy\" output: ");
      if(total_energy_output_obj_ptr==NULL) {
	msg += string("PROGRAMMING ERROR:"
		      " total_energy_output_obj_ptr not set."
		      " Driver Init() probably not called.");
      } else {
	msg += string(Tcl_GetStringResult(mif_interp));
      }
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    BOOL err;
    old_energy = Nb_Atof(Tcl_GetStringResult(mif_interp),err);
    if(err) {
      string msg=string("Error detected in FillNewStageState method"
			" --- Invalid \"Total energy\" output: ");
      msg += string(Tcl_GetStringResult(mif_interp));
      throw Oxs_Ext::Error(this,msg.c_str());
    }
  }
  new_state.AddDerivedData("Last energy",old_energy);
}

BOOL
Oxs_MinDriver::Step
(Oxs_ConstKey<Oxs_SimState> base_state,
 Oxs_Key<Oxs_SimState>& next_state)
{ // Returns true if step was successful, false if
  // unable to step as requested.

  // Put write lock on evolver in order to get a non-const
  // pointer.  Use a temporary variable, temp_key, so
  // write lock is automatically removed when temp_key
  // is destroyed.
  Oxs_Key<Oxs_MinEvolver> temp_key = evolver_key;
  Oxs_MinEvolver& evolver = temp_key.GetWriteReference();
  return evolver.Step(this,base_state,next_state);
}
