/* FILE: rectangularmesh.h            -*-Mode: c++-*-
 *
 * Cell rectangular mesh, derived from Oxs_Mesh class.
 *
 */

#ifndef _OXS_RECTANGULARMESH
#define _OXS_RECTANGULARMESH

#include <string>
#include <vector>

#include "atlas.h"
#include "scalarfield.h"
#include "mesh.h"

/* End includes */


class Oxs_RectangularMesh : public Oxs_Mesh {
private:
  ThreeVector base,cellsize;
  REAL8m cellvolume; // Convenience value
  UINT4m xdim,ydim,zdim;
  UINT4m xydim; // Convenience value: xdim*ydim
  UINT4m elementcount;  // Convenience value: xdim*ydim*zdim

  BOOL GetNeighborPoint(const ThreeVector& base_pt,
			UINT4m ngbr_index,
			ThreeVector& ngbr_pt) const;
  // Fills ngbr_pt with location of neighbor element indexed
  // by "ngbr_index", relative to base_pt.  Returns 1 if
  // ngbr_pt < number of neighbors (currently 6); otherwise
  // 0 is returned, in which case ngbr_pt is unchanged.
  // NB: ngbr_index is 0-based.

public:
  // Oxs_Ext interface
  Oxs_RectangularMesh
  (const char* name,     // Child instance id
   Oxs_Director* newdtr, // App director
   const char* argstr);  // MIF input block parameters
  virtual ~Oxs_RectangularMesh() {}
  virtual BOOL Init() { return 1; }
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.

  // Mesh interface
  virtual void GetBoundingBox(Oxs_Box& bbox) const;
  virtual UINT4m Size() const { return elementcount; }
  virtual REAL8m TotalVolume() const { return cellvolume*elementcount; }
  virtual REAL8m Volume(UINT4m) const { return cellvolume; }
  virtual void Center(UINT4m index,ThreeVector& location) const;


  // File output.  Throws an exception on error.
  virtual void WriteOvf
  (Tcl_Channel channel,   // Output channel
   const char* title,     // Long filename or title
   const char* desc,      // Description to embed in output file
   const char* valueunit, // Field units, such as "A/m".
   const char* meshtype,  // Either "rectangular" or "irregular"
   const char* datatype,  // Either "binary" or "text"
   const char* precision, // For binary, "4" or "8";
                          ///  for text, a printf-style format
   const Oxs_MeshValue<ThreeVector>* vec,  // Vector array
   const Oxs_MeshValue<REAL8m>* scale=NULL // Optional scaling for vec
   /// Set scale to NULL to use vec values directly.
   ) const;

  // Summing routines.  The terms are weighted by the basis element volumes.
  virtual REAL8m VolumeSum(const Oxs_MeshValue<REAL8m>& scalar) const;
  virtual ThreeVector VolumeSum(const Oxs_MeshValue<ThreeVector>& vec) const;
  virtual ThreeVector VolumeSum(const Oxs_MeshValue<ThreeVector>& vec,
				const Oxs_MeshValue<REAL8m>& scale) const;
  /// "scale" is elementwise scaling.

  // Extra precision summing routines.  These are the same as above,
  // except that intermediate sums are accumulated into Nb_Xpfloat
  // variables.
  virtual REAL8m VolumeSumXp(const Oxs_MeshValue<REAL8m>& scalar) const;
  virtual ThreeVector VolumeSumXp(const Oxs_MeshValue<ThreeVector>& vec) const;
  virtual ThreeVector VolumeSumXp(const Oxs_MeshValue<ThreeVector>& vec,
				  const Oxs_MeshValue<REAL8m>& scale) const;

  // Max angle routine.  Every mesh should have some concept of a
  // "neighborhood."  This routine should return the minimum dot
  // product between neighboring vectors across the mesh.  If the
  // input Oxs_MeshValue<ThreeVector> array are unit vectors, then
  // the max angle can be construed from acos(mindot).
  // NB: This routine is boundary condition sensitive.
  virtual REAL8m MinNeighborDot(const Oxs_MeshValue<ThreeVector>& vec,
				const Oxs_MeshValue<REAL8m>& zero_check)
    const;

  // Rectangular mesh specific functions.
  // IMPORTANT NOTE: Client routines can assume that the storage
  // order is first on x, second on y, third on z (Fortran order),
  // i.e., (0,0,0) (1,0,0) ... (xdim-1,0,0) (0,1,0) (1,1,0) ...
  // (xdim-1,ydim-1,0) (0,0,1) (1,0,1) ... (xdim-1,ydim-1,zdim-1)
  // This locks down the representation in this function, but
  // allows improved access in the clients.  If we need to change
  // representations, we can always introduce a new mesh class. ;^)
  REAL8m EdgeLengthX() const { return cellsize.x; }
  REAL8m EdgeLengthY() const { return cellsize.y; }
  REAL8m EdgeLengthZ() const { return cellsize.z; }
  UINT4m DimX() const { return xdim; }
  UINT4m DimY() const { return ydim; }
  UINT4m DimZ() const { return zdim; }


  // Boundary extraction.  Returns list (technically, an STL vector) of
  // indices for those elements inside base_region that have a neighbor
  // (in the 6 nearest ngbr sense) such that the first element lies on
  // one side (the "inside") of the surface specified by the
  // Oxs_ScalarField bdry_surface + bdry_value, and the neighbor lies on
  // the other (the "outside").  If the bdry_side argument is "-", then
  // the "inside" of the surface is the set of those points x for which
  // bdry_surface.Value(x)<bdry_value.  If bdry_side is "+", then the
  // "inside" of the surface is the set of those points x for which
  // bdry_surface.Value(x)>bdry_value.
  // Return value is the number of entries in the export list.
  // NB: The tested neighbors may lie outside the mesh proper, allowing
  // elements on the edge of the mesh to be specified.
  UINT4m BoundaryList(const Oxs_Atlas& atlas,
		      const string& region_name,
		      const Oxs_ScalarField& bdry_surface,
		      REAL8m bdry_value,
		      const string& bdry_side,
		      vector<UINT4m> &BoundaryIndexList) const;

  // Use the following function to convert a triplet index (x,y,z)
  // to the corresponding address used to access linear MeshValue
  // arrays.
  inline UINT4m Index(UINT4m x,UINT4m y,UINT4m z) const {
    return x+y*xdim+z*xydim;
  }

};

#endif // _OXS_RECTANGULARMESH
