/* FILE: transformzeeman.h         -*-Mode: c++-*-
 *
 * Zeeman (applied field) energy, derived from Oxs_Energy class,
 * specified from a Tcl proc.  This is a generalization of the
 * Oxs_ScriptUZeeman and Oxs_FixedZeeman classes.
 *
 */

#include "oc.h"
#include "nb.h"
#include "transformzeeman.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_TransformZeeman);

/* End includes */


// Constructor
Oxs_TransformZeeman::Oxs_TransformZeeman(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr), hmult(1.0),
    transform_type(identity), mesh_id(0)
{
  // Process arguments
  hmult = GetRealInitValue("multiplier",1.0);

  string xform_type_string = GetStringInitValue("type","identity");
  if(xform_type_string.compare("identity")==0) {
    transform_type = identity;
  } else if(xform_type_string.compare("diagonal")==0) {
    transform_type = diagonal;
  } else if(xform_type_string.compare("symmetric")==0) {
    transform_type = symmetric;
  } else if(xform_type_string.compare("general")==0) {
    transform_type = general;
  } else {
    string msg = string("Invalid transform type: \"")
      + xform_type_string
      + string("\".  Should be identity, diagonal, symmetric or general.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  if(transform_type != identity) {
    string cmdoptreq = GetStringInitValue("script_args",
					  "stage stage_time total_time");
    command_options.push_back(Oxs_TclCommandLineOption("stage",1));
    command_options.push_back(Oxs_TclCommandLineOption("stage_time",1));
    command_options.push_back(Oxs_TclCommandLineOption("total_time",1));
    cmd.SetBaseCommand(InstanceName(),
		       director->GetMifInterp(),
		       GetStringInitValue("script"),
		       Oxs_ParseTclCommandLineRequest(InstanceName(),
						      command_options,
						      cmdoptreq));
  }

  OXS_GET_INIT_EXT_OBJECT("field",Oxs_VectorField,fixedfield_init);

  VerifyAllInitArgsUsed();
}

Oxs_TransformZeeman::~Oxs_TransformZeeman()
{}

BOOL Oxs_TransformZeeman::Init()
{
  return Oxs_Energy::Init();
}

void
Oxs_TransformZeeman::GetAppliedField
(const Oxs_SimState& state,
 ThreeVector& row1, ThreeVector& row2, ThreeVector& row3,
 ThreeVector& drow1, ThreeVector& drow2, ThreeVector& drow3
) const
{
  // Special handling for identity transform (no script)
  if(transform_type == identity) {
    row1.Set(1,0,0); row2.Set(0,1,0); row3.Set(0,0,1);
    drow1.Set(0,0,0); drow2.Set(0,0,0); drow3.Set(0,0,0);
    return;
  }

  // Run script
  cmd.SaveInterpResult();

  int index;
  if((index = command_options[0].position)>=0) { // stage
    cmd.SetCommandArg(index,state.stage_number);
  }
  if((index = command_options[1].position)>=0) { // stage_time
    cmd.SetCommandArg(index,state.stage_elapsed_time);
  }
  if((index = command_options[2].position)>=0) { // total_time
    cmd.SetCommandArg(index,
		      state.stage_start_time+state.stage_elapsed_time);
  }

  cmd.Eval();

  if(cmd.GetResultListSize()>18) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Script return list too large: %d values",
		cmd.GetResultListSize());
    cmd.DiscardInterpResult();
    throw Oxs_Ext::Error(this,buf);
  }

  // Convert script string values to REAL8m's.
  vector<REAL8m> srv; // "Script Return Value"
  for(int i=0;i<cmd.GetResultListSize();i++) {
    REAL8m value;
    cmd.GetResultListItem(i,value);
    srv.push_back(value);
  }

  // Fill row vectors
  switch(transform_type) {
  case diagonal:
    if(srv.size()!=6) {
      string msg
	= string("Error detected during field evaluation.")
	+ string(" Script return is not a 6-tuple: ")
	+ cmd.GetWholeResult();
      cmd.DiscardInterpResult();
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    row1.Set(srv[0],0,0);
    row2.Set(0,srv[1],0);
    row3.Set(0,0,srv[2]);
    drow1.Set(srv[3],0,0);
    drow2.Set(0,srv[4],0);
    drow3.Set(0,0,srv[5]);
    break;
  case symmetric:
    if(srv.size()!=12) {
      string msg
	= string("Error detected during field evaluation.")
	+ string(" Script return is not a 12-tuple: ")
	+ cmd.GetWholeResult();
      cmd.DiscardInterpResult();
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    row1.Set(srv[0],srv[1],srv[2]);
    row2.Set(srv[1],srv[3],srv[4]);
    row3.Set(srv[2],srv[4],srv[5]);
    drow1.Set(srv[6],srv[7],srv[8]);
    drow2.Set(srv[7],srv[9],srv[10]);
    drow3.Set(srv[8],srv[10],srv[11]);
    break;
  case general:
    if(srv.size()!=18) {
      string msg
	= string("Error detected during field evaluation.")
	+ string(" Script return is not a 18-tuple: ")
	+ cmd.GetWholeResult();
      cmd.DiscardInterpResult();
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    row1.Set(srv[0],srv[1],srv[2]);
    row2.Set(srv[3],srv[4],srv[5]);
    row3.Set(srv[6],srv[7],srv[8]);
    drow1.Set(srv[9],srv[10],srv[11]);
    drow2.Set(srv[12],srv[13],srv[14]);
    drow3.Set(srv[15],srv[16],srv[17]);
    break;
  default:
    cmd.DiscardInterpResult();
    throw Oxs_Ext::Error(this,"Programming error; invalid transform type");
  }
  cmd.RestoreInterpResult();
}


void Oxs_TransformZeeman::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  UINT4m size = state.mesh->Size();
  if(size<1) return; // Nothing to do!

  if(mesh_id != state.mesh->Id()) {
    // This is either the first pass through, or else mesh
    // has changed.
    mesh_id = 0;
    fixedfield_init->FillMeshValue(state.mesh,fixedfield);
    if(hmult!=1.0) {
      for(UINT4m i=0;i<size;i++) fixedfield[i] *= hmult;
    }
    mesh_id = state.mesh->Id();
  }

  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  const Oxs_Mesh* mesh = state.mesh;

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  if(transform_type == identity) {
    UINT4m i=0;
    do {
      field[i] = fixedfield[i];
      energy[i] = -MU0*Ms[i]*(fixedfield[i]*spin[i]);
      ++i;
    } while(i<size);
    oed.pE_pt = 0.0;
  } else {
    ThreeVector row1, row2, row3;
    ThreeVector drow1, drow2, drow3;
    GetAppliedField(state,row1,row2,row3,drow1,drow2,drow3);

    REAL8m pE_pt_sum=0;
    UINT4m i=0;
    do {
      const ThreeVector& v = fixedfield[i];
      field[i].Set(row1*v,row2*v,row3*v); // Apply transform

      ThreeVector temp = (-MU0*Ms[i])*spin[i];
      energy[i] = field[i]*temp;

      ThreeVector dH(drow1*v,drow2*v,drow3*v); // dH/dt
      pE_pt_sum += mesh->Volume(i)*(dH*temp);

      ++i;
    } while(i<size);
    oed.pE_pt = pE_pt_sum;
  }

}
