/* FILE: twosurfaceexchange.cc            -*-Mode: c++-*-
 *
 * Long range (RKKY-style) exchange energy on links across spacer
 * layer, with surface (interfacial) exchange energy coefficents
 * sigma (bilinear) and sigma2 (biquadratic).
 *
 */

#include "nb.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "twosurfaceexchange.h"
#include "energy.h"             // Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_TwoSurfaceExchange);

/* End includes */


// Constructor
Oxs_TwoSurfaceExchange::Oxs_TwoSurfaceExchange(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr),
    mesh_id(0)
{
  // Process arguments
  init_sigma = GetRealInitValue("sigma",0.);
  init_sigma2 = GetRealInitValue("sigma2",0.);

  // First boundary.  Import parameter string should be a 10 element
  // list of name-value pairs:
  //         NAME         VALUE
  //         atlas        atlas reference
  //         region       region name
  //         scalarfield  scalar field reference
  //         scalarvalue  scalar field level surface reference value
  //         scalarside   either "+" or "-", denoting inner surface side
  CheckInitValueParamCount("surface1",10);
  vector<string> surface1;
  FindRequiredInitValue("surface1",surface1);
  BOOL badinput1=0;
  if(surface1[0].compare("atlas")       != 0  ||
     surface1[2].compare("region")      != 0  ||
     surface1[4].compare("scalarfield") != 0  ||
     surface1[6].compare("scalarvalue") != 0  || 
     surface1[8].compare("scalarside")  != 0    ) {
    badinput1 = 1;
  } else {
    Oxs_SplitList param_strings;
    vector<string> params;
    param_strings.Split(surface1[1].c_str());
    param_strings.FillParams(params);
    OXS_GET_EXT_OBJECT(params,Oxs_Atlas,atlas1);
    region1 = surface1[3];
    param_strings.Split(surface1[5].c_str());
    param_strings.FillParams(params);
    OXS_GET_EXT_OBJECT(params,Oxs_ScalarField,bdry1);
    bdry1_value = Nb_Atof(surface1[7].c_str());
    bdry1_side = surface1[9];
    if(atlas1->GetRegionId(region1) == -1 ||
       (bdry1_side.compare("+")!=0 && bdry1_side.compare("-")!=0)) {
      badinput1 = 1;
    }
  }
  DeleteInitValue("surface1");

  // Second boundary.  Import parameter string same format
  // as for first boundary.
  CheckInitValueParamCount("surface2",10);
  vector<string> surface2;
  FindRequiredInitValue("surface2",surface2);
  BOOL badinput2=0;
  if(surface2[0].compare("atlas")       != 0  ||
     surface2[2].compare("region")      != 0  ||
     surface2[4].compare("scalarfield") != 0  ||
     surface2[6].compare("scalarvalue") != 0  || 
     surface2[8].compare("scalarside")  != 0    ) {
    badinput2 = 1;
  } else {
    Oxs_SplitList param_strings;
    vector<string> params;
    param_strings.Split(surface2[1].c_str());
    param_strings.FillParams(params);
    OXS_GET_EXT_OBJECT(params,Oxs_Atlas,atlas2);
    region2 = surface2[3];
    param_strings.Split(surface2[5].c_str());
    param_strings.FillParams(params);
    OXS_GET_EXT_OBJECT(params,Oxs_ScalarField,bdry2);
    bdry2_value = Nb_Atof(surface2[7].c_str());
    bdry2_side = surface2[9];
    if(atlas2->GetRegionId(region2) == -1 ||
       (bdry2_side.compare("+")!=0 && bdry2_side.compare("-")!=0)) {
      badinput2 = 1;
    }
  }
  DeleteInitValue("surface2");

  if(badinput1 || badinput2) {
    string msg=
      string("Bad surface parameter block, inside "
             "Oxs_TwoSurfaceExchange mif Specify block for object ")
      +  string(InstanceName()) + string(":");
    if(badinput1) msg += string(" surface1 block is bad");
    if(badinput1 && badinput2) msg += string(", ");
    if(badinput2) msg += string(" surface2 block is bad");
    msg += string(".");
    throw Oxs_Ext::Error(msg.c_str());
  }

  VerifyAllInitArgsUsed();
}

Oxs_TwoSurfaceExchange::~Oxs_TwoSurfaceExchange()
{}

void Oxs_TwoSurfaceExchange::FillLinkList
(const Oxs_RectangularMesh* mesh) const
{ // Queries the mesh with each surface spec,
  // and then pairs up cells between surfaces.
  // Each cell on the first surface is paired up with
  // the closest cell from the second surface.  Note
  // the asymmetry in this pairing procedure---generally,
  // one will probably want to make the smaller surface
  // the "first" surface.  Consider these pictures:
  //
  //          22222222222222 <-- Surface 2
  //              ||||||           
  //              ||||||  <-- Links
  //              ||||||         
  //              111111  <-- Surface 1
  //
  //  as compared to
  //
  //          11111111111111 <-- Surface 1
  //           \\\||||||///        
  //            \\||||||// <-- Links
  //             \||||||/        
  //              222222  <-- Surface 2
  //
  // In the second picture, all the outer cells in the top
  // surface get paired up with the outer elements in the
  // lower surface.

  // Get cell lists for each surface
  vector<UINT4m> bdry1_cells,bdry2_cells;
  mesh->BoundaryList(*atlas1,region1,*bdry1,bdry1_value,bdry1_side,
                     bdry1_cells);
  mesh->BoundaryList(*atlas2,region2,*bdry2,bdry2_value,bdry2_side,
                     bdry2_cells);
  if(bdry1_cells.size()<1 || bdry2_cells.size()<1) {
    string msg =
      string("Empty bdry/surface list with current mesh"
             " detected in Oxs_TwoSurfaceExchange::FillLinkList()"
             " routine of object ") + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  // For each cell in bdry1_cells, find closest cell in bdry2_cells,
  // and store this pair in "links".
  links.clear();  // Throw away previous links, if any.
  vector<UINT4m>::iterator it1 = bdry1_cells.begin();
  while(it1!=bdry1_cells.end()) {
    ThreeVector cell1;
    mesh->Center(*it1,cell1);
    vector<UINT4m>::iterator it2 = bdry2_cells.begin();
    ThreeVector vtemp;
    mesh->Center(*it2,vtemp);
    vtemp -= cell1;
    REAL8m min_distsq = vtemp.MagSq();
    UINT4m min_index = *it2;
    while( (++it2) != bdry2_cells.end()) {
      mesh->Center(*it2,vtemp);
      vtemp -= cell1;
      REAL8m distsq = vtemp.MagSq();
      if(distsq<min_distsq) {
        min_distsq = distsq;
        min_index = *it2;
      }
    }
    if(min_distsq>0.) {
      // The sigma coef's need to be weighted by the inverse
      // of the thickness of the compuational cell in the
      // direction of the link.  This is necessary so that
      // the total surface (interfacial) exchange energy is
      // independent of the discretization size, as the
      // discretization size is reduces to 0.  (As the
      // computational cell size is reduced, the surface
      // energy is squeezed into an ever thinner layer of
      // cells along the surface.)
      ThreeVector vtemp;
      mesh->Center(min_index,vtemp);
      vtemp -= cell1;
      REAL8m cellwgt = vtemp.x*mesh->EdgeLengthX()
                     + vtemp.y*mesh->EdgeLengthY()
                     + vtemp.z*mesh->EdgeLengthZ();
      if(cellwgt>0.) { // Safety
	cellwgt = sqrt(min_distsq)/cellwgt;
	// min_distsq _should_ be vtemp.MagSq(), so
	// cellwgt is |vtemp|/<vtemp,celldims>, i.e.,
	// < vtemp/|vtemp| , celldims >^{-1}
      }

      Oxs_TwoSurfaceExchangeLinkParams ltemp;
      ltemp.index1 = *it1;    ltemp.index2 = min_index;
      ltemp.exch_coef1 = (init_sigma + init_sigma2) * cellwgt;
      ltemp.exch_coef2 = -init_sigma2 * cellwgt;
      // NB: The init_sigma and init_sigma2 references above
      //  will likely be replaced in the future with a call
      //  to a user-defined function taking as imports the
      //  center locations of the two cells in the link pair.

      links.push_back(ltemp);
    }
    ++it1;
  }
}

void Oxs_TwoSurfaceExchange::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    string msg="Import mesh to Oxs_TwoSurfaceExchange::GetEnergy()"
      " routine of object " +  string(InstanceName())
      + " is not an Oxs_RectangularMesh object.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  // If mesh has changed, re-pick link selections
  if(mesh_id !=  mesh->Id()) {
    FillLinkList(mesh);
    mesh_id=mesh->Id();
  }

  // Zero entire energy and field meshes
  UINT4m size=mesh->Size();
  UINT4m index;
  for(index=0;index<size;index++) energy[index]=0.;
  for(index=0;index<size;index++) field[index].Set(0.,0.,0.);

  // Iterate through link list and accumulate energies and fields
  REAL8m hcoef = 2.0/MU0;
  vector<Oxs_TwoSurfaceExchangeLinkParams>::const_iterator it
    = links.begin();
  for(it=links.begin();it!=links.end();++it) {
    UINT4m i = it->index1;
    UINT4m j = it->index2;
    if(Ms_inverse[i]==0. || Ms_inverse[j]==0.) continue; // Ms=0; skip
    REAL8m ecoef1 = it->exch_coef1;
    REAL8m ecoef2 = it->exch_coef2;
    ThreeVector mdiff = spin[i]-spin[j];
    REAL8m temp = 0.5*mdiff.MagSq();
    REAL8m elink = (ecoef1 + ecoef2*temp)*temp; // Energy density
    energy[i] += elink;
    energy[j] += elink;
    mdiff *= hcoef*(ecoef1+2*ecoef2*temp);
    field[i] += -1*Ms_inverse[i]*mdiff;
    field[j] +=    Ms_inverse[j]*mdiff;
  }
}

