/* FILE: uniformexchange.cc            -*-Mode: c++-*-
 *
 * Uniform 6 neighbor exchange energy on rectangular mesh,
 * derived from Oxs_Energy class.
 *
 */

#include "nb.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "uniformexchange.h"
#include "energy.h"             // Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_UniformExchange);

/* End includes */


// Constructor
Oxs_UniformExchange::Oxs_UniformExchange(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr),
    A(0.), kernel(UNKNOWN)
{
  // Process arguments
  A = GetRealInitValue("A");
  string kernel_request = GetStringInitValue("kernel","6ngbr");
  if(kernel_request.compare("6ngbr")==0) {
    kernel = NGBR_6;
  } else if(kernel_request.compare("12ngbr")==0) {
    kernel = NGBR_12;
  } else if(kernel_request.compare("12ngbrmirror")==0) {
    kernel = NGBR_12_MIRROR;
  } else if(kernel_request.compare("26ngbr")==0) {
    kernel = NGBR_26;
  } else {
    string msg=string("Invalid kernel request: ")
      + kernel_request
      + string("\n Should be one of 6ngbr, 12ngbr, 12ngbrmirror, or 26ngbr.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  VerifyAllInitArgsUsed();
}

Oxs_UniformExchange::~Oxs_UniformExchange()
{}

BOOL Oxs_UniformExchange::Init()
{
  return Oxs_Energy::Init();
}

void
Oxs_UniformExchange::CalcEnergy6Ngbr
(const Oxs_MeshValue<ThreeVector>& spin,
 const Oxs_MeshValue<REAL8m>& Ms_inverse,
 const Oxs_RectangularMesh* mesh,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field) const
{
  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  const REAL8m hcoef = -2/MU0;
  REAL8m wgtx = -A/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = -A/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = -A/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
        UINT4m i = mesh->Index(x,y,z); // Get base linear address

        REAL8m Msii = Ms_inverse[i];
        if(Msii == 0.0) {
          energy[i]=0.0;
          field[i].Set(0.,0.,0.);
          continue;
        }

        ThreeVector base = spin[i];

        ThreeVector sum(0.,0.,0.);
        if(x>0) {
          UINT4m j = i-1;
          if(Ms_inverse[j]!=0.0) sum = (spin[j] - base);
        }
        if(x<xdim-1) {
          UINT4m j = i+1;
          if(Ms_inverse[j]!=0.0) {
            sum += (spin[j] - base);
          }
        }
        sum *= wgtx;

        ThreeVector temp(0.,0.,0.);
        if(y>0) {
          UINT4m j = i-xdim;
          if(Ms_inverse[j]!=0.0) temp = (spin[j] - base);
        }
        if(y<ydim-1) {
          UINT4m j = i+xdim;
          if(Ms_inverse[j]!=0.0) {
            temp += (spin[j] - base);
          }
        }
        sum += wgty*temp;

        temp.Set(0.,0.,0.);
        if(z>0) {
          UINT4m j = i-xydim;
          if(Ms_inverse[j]!=0.0) temp = (spin[j] - base);
        }
        if(z<zdim-1) {
          UINT4m j = i+xydim;
          if(Ms_inverse[j]!=0.0) {
            temp += (spin[j]- base);
          }
        }
        sum += wgtz*temp;

        field[i] = (hcoef*Msii) * sum;
        energy[i] = (sum * base);
      }
    }
  }
}

void
Oxs_UniformExchange::CalcEnergy12Ngbr
(const Oxs_MeshValue<ThreeVector>& spin,
 const Oxs_MeshValue<REAL8m>& Ms_inverse,
 const Oxs_RectangularMesh* mesh,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field) const
{
  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  const REAL8m hcoef = -2/MU0;
  REAL8m wgtx = -A/(mesh->EdgeLengthX()*mesh->EdgeLengthX()*1152.);
  REAL8m wgty = -A/(mesh->EdgeLengthY()*mesh->EdgeLengthY()*1152.);
  REAL8m wgtz = -A/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ()*1152.);

  // See mjd's NOTES II, 12-Aug-2002, p186-192
  // EXTRA NOTE: Beware that "field" computed here is
  //  -grad(E)/(mu_0*cellvolume), which is not a good representation
  //  for H.  See mjd's NOTES II, 9-Aug-2002, p183.
  if(xdim<10 || ydim<10 || zdim<10) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"All dimensions must be at least 10 for 12ngbr kernel."
		" (Actual dimensions: xdim=%d, ydim=%d, zdim=%d.)",
		xdim,ydim,zdim);
    throw Oxs_Ext::Error(this,buf);
  }
  // Note: Integral weights ai, bj, and ck are derived in mjd
  // NOTES II, 5-Aug-2002, p178-181.
  for(UINT4m z=0;z<zdim;z++) {
    REAL8m ck=1.0;
    if(z==0 || z==zdim-1) ck=26./24.;
    if(z==1 || z==zdim-2) ck=21./24.;
    if(z==2 || z==zdim-3) ck=25./24.;

    for(UINT4m y=0;y<ydim;y++) {
      REAL8m bj=1.0;
      if(y==0 || y==ydim-1) bj=26./24.;
      if(y==1 || y==ydim-2) bj=21./24.;
      if(y==2 || y==ydim-3) bj=25./24.;

      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address

	REAL8m Msii = Ms_inverse[i];
	if(Msii == 0.0) {
	  energy[i]=0.0;
	  field[i].Set(0.,0.,0.);
	  continue;
	}

	REAL8m ai=1.0;
	if(x==0 || x==xdim-1) ai=26./24.;
	if(x==1 || x==xdim-2) ai=21./24.;
	if(x==2 || x==xdim-3) ai=25./24.;

	ThreeVector sum,tsum;

	// Dxx
	if(x>4 && x<xdim-5) {
	  tsum  =    -96*spin[i-2];
	  tsum +=   1536*spin[i-1];
	  tsum +=  -2880*spin[i];
	  tsum +=   1536*spin[i+1];
	  tsum +=    -96*spin[i+2];
	} else if(x==0) {
	  tsum  =  -6125*spin[i];
	  tsum +=  11959*spin[i+1];
	  tsum +=  -8864*spin[i+2];
	  tsum +=   3613*spin[i+3];
	  tsum +=   -583*spin[i+4];
	} else if(x==1) {
	  tsum  =  11959*spin[i-1];
	  tsum += -25725*spin[i];
	  tsum +=  20078*spin[i+1];
	  tsum +=  -7425*spin[i+2];
	  tsum +=   1113*spin[i+3];
	} else if(x==2) {
	  tsum  =  -8864*spin[i-2];
	  tsum +=  20078*spin[i-1];
	  tsum += -17175*spin[i];
	  tsum +=   6752*spin[i+1];
	  tsum +=   -791*spin[i+2];
	} else if(x==3) {
	  tsum  =   3613*spin[i-3];
	  tsum +=  -7425*spin[i-2];
	  tsum +=   6752*spin[i-1];
	  tsum +=  -4545*spin[i];
	  tsum +=   1701*spin[i+1];
	  tsum +=    -96*spin[i+2];
	} else if(x==4) {
	  tsum  =   -583*spin[i-4];
	  tsum +=   1113*spin[i-3];
	  tsum +=   -791*spin[i-2];
	  tsum +=   1701*spin[i-1];
	  tsum +=  -2880*spin[i];
	  tsum +=   1536*spin[i+1];
	  tsum +=    -96*spin[i+2];
	} else if(x==xdim-5) {
	  tsum  =    -96*spin[i-2];
	  tsum +=   1536*spin[i-1];
	  tsum +=  -2880*spin[i];
	  tsum +=   1701*spin[i+1];
	  tsum +=   -791*spin[i+2];
	  tsum +=   1113*spin[i+3];
	  tsum +=   -583*spin[i+4];
	} else if(x==xdim-4) {
	  tsum  =    -96*spin[i-2];
	  tsum +=   1701*spin[i-1];
	  tsum +=  -4545*spin[i];
	  tsum +=   6752*spin[i+1];
	  tsum +=  -7425*spin[i+2];
	  tsum +=   3613*spin[i+3];
	} else if(x==xdim-3) {
	  tsum  =   -791*spin[i-2];
	  tsum +=   6752*spin[i-1];
	  tsum += -17175*spin[i];
	  tsum +=  20078*spin[i+1];
	  tsum +=  -8864*spin[i+2];
	} else if(x==xdim-2) {
	  tsum  =   1113*spin[i-3];
	  tsum +=  -7425*spin[i-2];
	  tsum +=  20078*spin[i-1];
	  tsum += -25725*spin[i];
	  tsum +=  11959*spin[i+1];
	} else if(x==xdim-1) {
	  tsum  =   -583*spin[i-4];
	  tsum +=   3613*spin[i-3];
	  tsum +=  -8864*spin[i-2];
	  tsum +=  11959*spin[i-1];
	  tsum +=  -6125*spin[i];
	}
	sum = (bj*ck*wgtx)*tsum; // ai built into above coefficients.

	// Dyy
	if(y>4 && y<ydim-5) {
	  tsum  =    -96*spin[i-2*xdim];
	  tsum +=   1536*spin[i-xdim];
	  tsum +=  -2880*spin[i];
	  tsum +=   1536*spin[i+xdim];
	  tsum +=    -96*spin[i+2*xdim];
	} else if(y==0) {
	  tsum  =  -6125*spin[i];
	  tsum +=  11959*spin[i+xdim];
	  tsum +=  -8864*spin[i+2*xdim];
	  tsum +=   3613*spin[i+3*xdim];
	  tsum +=   -583*spin[i+4*xdim];
	} else if(y==1) {
	  tsum  =  11959*spin[i-xdim];
	  tsum += -25725*spin[i];
	  tsum +=  20078*spin[i+xdim];
	  tsum +=  -7425*spin[i+2*xdim];
	  tsum +=   1113*spin[i+3*xdim];
	} else if(y==2) {
	  tsum  =  -8864*spin[i-2*xdim];
	  tsum +=  20078*spin[i-xdim];
	  tsum += -17175*spin[i];
	  tsum +=   6752*spin[i+xdim];
	  tsum +=   -791*spin[i+2*xdim];
	} else if(y==3) {
	  tsum  =   3613*spin[i-3*xdim];
	  tsum +=  -7425*spin[i-2*xdim];
	  tsum +=   6752*spin[i-xdim];
	  tsum +=  -4545*spin[i];
	  tsum +=   1701*spin[i+xdim];
	  tsum +=    -96*spin[i+2*xdim];
	} else if(y==4) {
	  tsum  =   -583*spin[i-4*xdim];
	  tsum +=   1113*spin[i-3*xdim];
	  tsum +=   -791*spin[i-2*xdim];
	  tsum +=   1701*spin[i-xdim];
	  tsum +=  -2880*spin[i];
	  tsum +=   1536*spin[i+xdim];
	  tsum +=    -96*spin[i+2*xdim];
	} else if(y==ydim-5) {
	  tsum  =    -96*spin[i-2*xdim];
	  tsum +=   1536*spin[i-xdim];
	  tsum +=  -2880*spin[i];
	  tsum +=   1701*spin[i+xdim];
	  tsum +=   -791*spin[i+2*xdim];
	  tsum +=   1113*spin[i+3*xdim];
	  tsum +=   -583*spin[i+4*xdim];
	} else if(y==ydim-4) {
	  tsum  =    -96*spin[i-2*xdim];
	  tsum +=   1701*spin[i-xdim];
	  tsum +=  -4545*spin[i];
	  tsum +=   6752*spin[i+xdim];
	  tsum +=  -7425*spin[i+2*xdim];
	  tsum +=   3613*spin[i+3*xdim];
	} else if(y==ydim-3) {
	  tsum  =   -791*spin[i-2*xdim];
	  tsum +=   6752*spin[i-xdim];
	  tsum += -17175*spin[i];
	  tsum +=  20078*spin[i+xdim];
	  tsum +=  -8864*spin[i+2*xdim];
	} else if(y==ydim-2) {
	  tsum  =   1113*spin[i-3*xdim];
	  tsum +=  -7425*spin[i-2*xdim];
	  tsum +=  20078*spin[i-xdim];
	  tsum += -25725*spin[i];
	  tsum +=  11959*spin[i+xdim];
	} else if(y==ydim-1) {
	  tsum  =   -583*spin[i-4*xdim];
	  tsum +=   3613*spin[i-3*xdim];
	  tsum +=  -8864*spin[i-2*xdim];
	  tsum +=  11959*spin[i-xdim];
	  tsum +=  -6125*spin[i];
	}
	sum += (ai*ck*wgty)*tsum; // bj built into above coefficients.

	// Dzz
	if(z>4 && z<zdim-5) {
	  tsum  =    -96*spin[i-2*xydim];
	  tsum +=   1536*spin[i-xydim];
	  tsum +=  -2880*spin[i];
	  tsum +=   1536*spin[i+xydim];
	  tsum +=    -96*spin[i+2*xydim];
	} else if(z==0) {
	  tsum  =  -6125*spin[i];
	  tsum +=  11959*spin[i+xydim];
	  tsum +=  -8864*spin[i+2*xydim];
	  tsum +=   3613*spin[i+3*xydim];
	  tsum +=   -583*spin[i+4*xydim];
	} else if(z==1) {
	  tsum  =  11959*spin[i-xydim];
	  tsum += -25725*spin[i];
	  tsum +=  20078*spin[i+xydim];
	  tsum +=  -7425*spin[i+2*xydim];
	  tsum +=   1113*spin[i+3*xydim];
	} else if(z==2) {
	  tsum  =  -8864*spin[i-2*xydim];
	  tsum +=  20078*spin[i-xydim];
	  tsum += -17175*spin[i];
	  tsum +=   6752*spin[i+xydim];
	  tsum +=   -791*spin[i+2*xydim];
	} else if(z==3) {
	  tsum  =   3613*spin[i-3*xydim];
	  tsum +=  -7425*spin[i-2*xydim];
	  tsum +=   6752*spin[i-xydim];
	  tsum +=  -4545*spin[i];
	  tsum +=   1701*spin[i+xydim];
	  tsum +=    -96*spin[i+2*xydim];
	} else if(z==4) {
	  tsum  =   -583*spin[i-4*xydim];
	  tsum +=   1113*spin[i-3*xydim];
	  tsum +=   -791*spin[i-2*xydim];
	  tsum +=   1701*spin[i-xydim];
	  tsum +=  -2880*spin[i];
	  tsum +=   1536*spin[i+xydim];
	  tsum +=    -96*spin[i+2*xydim];
	} else if(z==zdim-5) {
	  tsum  =    -96*spin[i-2*xydim];
	  tsum +=   1536*spin[i-xydim];
	  tsum +=  -2880*spin[i];
	  tsum +=   1701*spin[i+xydim];
	  tsum +=   -791*spin[i+2*xydim];
	  tsum +=   1113*spin[i+3*xydim];
	  tsum +=   -583*spin[i+4*xydim];
	} else if(z==zdim-4) {
	  tsum  =    -96*spin[i-2*xydim];
	  tsum +=   1701*spin[i-xydim];
	  tsum +=  -4545*spin[i];
	  tsum +=   6752*spin[i+xydim];
	  tsum +=  -7425*spin[i+2*xydim];
	  tsum +=   3613*spin[i+3*xydim];
	} else if(z==zdim-3) {
	  tsum  =   -791*spin[i-2*xydim];
	  tsum +=   6752*spin[i-xydim];
	  tsum += -17175*spin[i];
	  tsum +=  20078*spin[i+xydim];
	  tsum +=  -8864*spin[i+2*xydim];
	} else if(z==zdim-2) {
	  tsum  =   1113*spin[i-3*xydim];
	  tsum +=  -7425*spin[i-2*xydim];
	  tsum +=  20078*spin[i-xydim];
	  tsum += -25725*spin[i];
	  tsum +=  11959*spin[i+xydim];
	} else if(z==zdim-1) {
	  tsum  =   -583*spin[i-4*xydim];
	  tsum +=   3613*spin[i-3*xydim];
	  tsum +=  -8864*spin[i-2*xydim];
	  tsum +=  11959*spin[i-xydim];
	  tsum +=  -6125*spin[i];
	}
	sum += (ai*bj*wgtz)*tsum; // ck built into above coefficients.

	field[i] = (hcoef*Msii) * sum;
	energy[i] = (sum * spin[i]);
      }
    }
  }
}

void
Oxs_UniformExchange::CalcEnergy12NgbrMirror
(const Oxs_MeshValue<ThreeVector>& spin,
 const Oxs_MeshValue<REAL8m>& Ms_inverse,
 const Oxs_RectangularMesh* mesh,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field) const
{
  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  const REAL8m hcoef = -2/MU0;

  // Second derivative weighting matrix is (-1 16 -30 16 -1)/12.
  // Introduce 2 row mirror boundary, i.e., if edge is
  //
  //           |
  //           | a  b  c  d ...
  //           |
  // use
  //           |
  //      b  a | a  b  c  d ...
  //           |
  //
  REAL8m wgtx = -A/(mesh->EdgeLengthX()*mesh->EdgeLengthX()*12.);
  REAL8m wgty = -A/(mesh->EdgeLengthY()*mesh->EdgeLengthY()*12.);
  REAL8m wgtz = -A/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ()*12.);
  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address

	REAL8m Msii = Ms_inverse[i];
	if(Msii == 0.0) {
	  energy[i]=0.0;
	  field[i].Set(0.,0.,0.);
	  continue;
	}

	ThreeVector base = spin[i];

	ThreeVector sum_inner(0.,0.,0.);
	ThreeVector sum_outer(0.,0.,0.);
	if(x > 1) {
	  UINT4m j1 =  i - 1;
	  UINT4m j2 = j1 - 1;
	  if(Ms_inverse[j1]!=0.0) sum_inner = spin[j1] - base;
	  if(Ms_inverse[j2]!=0.0) sum_outer = spin[j1] - spin[j2];
	} else if(x == 1) {
	  UINT4m j = i - 1;
	  if(Ms_inverse[j]!=0.0) sum_inner = spin[j] - base;
	} else if(xdim > 1) {
	  // x==0.  Add edge correction.
	  UINT4m j = i + 1;
	  if(Ms_inverse[j]!=0.0) sum_outer = base - spin[j];
	}
	if(x+2 < xdim) {
	  // Note: We have to compare 'x+2<xdim', NOT 'x<xdim-1',
	  // because x and xdim are unsigned, and xdim may be
	  // as small as 1.
	  UINT4m j1 =  i + 1;
	  UINT4m j2 = j1 + 1;
	  if(Ms_inverse[j1]!=0.0) sum_inner += spin[j1] - base;
	  if(Ms_inverse[j2]!=0.0) sum_outer += spin[j1] - spin[j2];
	} else if(x+2 == xdim) {
	  UINT4m j = i + 1;
	  if(Ms_inverse[j]!=0.0) sum_inner += spin[j] - base;
	} else if(xdim > 1) {
	  // x==xdim-1.  Add edge correction.
	  UINT4m j = i - 1;
	  if(Ms_inverse[j]!=0.0) sum_outer += base - spin[j];
	}
	sum_inner *= wgtx;
	sum_outer *= wgtx;

	ThreeVector temp_inner(0.,0.,0.);
	ThreeVector temp_outer(0.,0.,0.);
	if(y > 1) {
	  UINT4m j1 =  i - xdim;
	  UINT4m j2 = j1 - xdim;
	  if(Ms_inverse[j1]!=0.0) temp_inner = spin[j1] - base;
	  if(Ms_inverse[j2]!=0.0) temp_outer = spin[j1] - spin[j2];
	} else if(y == 1) {
	  UINT4m j = i - xdim;
	  if(Ms_inverse[j]!=0.0) temp_inner = spin[j] - base;
	} else if(ydim > 1) {
	  // y==0.  Add edge correction.
	  UINT4m j = i + xdim;
	  if(Ms_inverse[j]!=0.0) temp_outer = base - spin[j];
	}
	if(y+2 < ydim) {
	  UINT4m j1 =  i + xdim;
	  UINT4m j2 = j1 + xdim;
	  if(Ms_inverse[j1]!=0.0) temp_inner += spin[j1] - base;
	  if(Ms_inverse[j2]!=0.0) temp_outer += spin[j1] - spin[j2];
	} else if(y+2 == ydim) {
	  UINT4m j = i + xdim;
	  if(Ms_inverse[j]!=0.0) temp_inner += spin[j] - base;
	} else if(ydim > 1) {
	  // y==ydim-1.  Add edge correction.
	  UINT4m j = i - xdim;
	  if(Ms_inverse[j]!=0.0) temp_outer += base - spin[j];
	}
	sum_inner += wgty*temp_inner;
	sum_outer += wgty*temp_outer;

	temp_inner.Set(0.,0.,0.);
	temp_outer.Set(0.,0.,0.);
	if(z > 1) {
	  UINT4m j1 =  i - xydim;
	  UINT4m j2 = j1 - xydim;
	  if(Ms_inverse[j1]!=0.0) temp_inner = spin[j1] - base;
	  if(Ms_inverse[j2]!=0.0) temp_outer = spin[j1] - spin[j2];
	} else if(z == 1) {
	  UINT4m j = i - xydim;
	  if(Ms_inverse[j]!=0.0) temp_inner = spin[j] - base;
	} else if(zdim > 1) {
	  // z==0.  Add edge correction.
	  UINT4m j = i + xydim;
	  if(Ms_inverse[j]!=0.0) temp_outer = base - spin[j];
	}
	if(z+2 < zdim) {
	  UINT4m j1 =  i + xydim;
	  UINT4m j2 = j1 + xydim;
	  if(Ms_inverse[j1]!=0.0) temp_inner += spin[j1] - base;
	  if(Ms_inverse[j2]!=0.0) temp_outer += spin[j1] - spin[j2];
	} else if(z+2 == zdim) {
	  UINT4m j = i + xydim;
	  if(Ms_inverse[j]!=0.0) temp_inner += spin[j] - base;
	} else if(zdim > 1) {
	  // z==zdim-1.  Add edge correction.
	  UINT4m j = i - xydim;
	  if(Ms_inverse[j]!=0.0) temp_outer += base - spin[j];
	}
	sum_inner += wgtz*temp_inner;
	sum_outer += wgtz*temp_outer;

	sum_inner *= 15;
	sum_inner += sum_outer;

	field[i] = (hcoef*Msii) * sum_inner;
	energy[i] = (sum_inner * base);
      }
    }
  }
}

void
Oxs_UniformExchange::CalcEnergy26Ngbr
(const Oxs_MeshValue<ThreeVector>& spin,
 const Oxs_MeshValue<REAL8m>& Ms_inverse,
 const Oxs_RectangularMesh* mesh,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field) const
{
  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  const REAL8m hcoef = -2/MU0;

  // Construct term weightings.  See mjd's NOTES I, 23-Oct-1997,
  // pp 168-171.
  REAL8m delxsq = mesh->EdgeLengthX()*mesh->EdgeLengthX();
  REAL8m delysq = mesh->EdgeLengthY()*mesh->EdgeLengthY();
  REAL8m delzsq = mesh->EdgeLengthZ()*mesh->EdgeLengthZ();
  REAL8m delisum = (1.0/delxsq)+(1.0/delysq)+(1.0/delzsq);
  REAL8m Amult = A/18.0;

  // Nearest neighbor weights.  These are all zero for cubic cells.
  // Might want to implement special handling in the algorithm
  // below for this common case.
  REAL8m wgt_x=0.,wgt_y=0.,wgt_z=0.;
  if(delxsq!=delysq || delxsq!=delzsq || delysq!=delzsq) {
    wgt_x = 4*Amult*(delisum-3.0/delxsq);
    wgt_y = 4*Amult*(delisum-3.0/delysq);
    wgt_z = 4*Amult*(delisum-3.0/delzsq);
  }

  // 2nd nearest neighbor weights.
  REAL8m wgt_xy = 2*Amult*(-delisum + 3.0/(2*delzsq));
  REAL8m wgt_xz = 2*Amult*(-delisum + 3.0/(2*delysq));
  REAL8m wgt_yz = 2*Amult*(-delisum + 3.0/(2*delxsq));

  // Corner neighbor weight.
  REAL8m wgt_corner = -0.5*Amult*delisum;

  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address

	REAL8m Msii = Ms_inverse[i];
	if(Msii == 0.0) {
	  energy[i]=0.0;
	  field[i].Set(0.,0.,0.);
	  continue;
	}

	ThreeVector base = spin[i];

	// Nearest x neighbors.  At most 2.
	ThreeVector sum_x(0.,0.,0.);
	if(x>0) {
	  UINT4m j = i-1;
	  if(Ms_inverse[j]!=0.0) sum_x = (spin[j] - base);
	}
	if(x<xdim-1) {
	  UINT4m j = i+1;
	  if(Ms_inverse[j]!=0.0) sum_x += (spin[j] - base);
	}
	sum_x *= wgt_x;

	// Nearest y neighbors.  At most 2.
	ThreeVector sum_y(0.,0.,0.);
	if(y>0) {
	  UINT4m j = i-xdim;
	  if(Ms_inverse[j]!=0.0) sum_y = (spin[j] - base);
	}
	if(y<ydim-1) {
	  UINT4m j = i+xdim;
	  if(Ms_inverse[j]!=0.0) sum_y += (spin[j] - base);
	}
	sum_y *= wgt_y;

	// Nearest z neighbors.  At most 2.
	ThreeVector sum_z(0.,0.,0.);
	if(z>0) {
	  UINT4m j = i-xydim;
	  if(Ms_inverse[j]!=0.0) sum_z = (spin[j] - base);
	}
	if(z<zdim-1) {
	  UINT4m j = i+xydim;
	  if(Ms_inverse[j]!=0.0) sum_z += (spin[j] - base);
	}
	sum_z *= wgt_z;

	// xy-neighbors.  At most 4.
	UINT4m j;
	ThreeVector sum_xy;
	j = i;  if(x>0)      j -= 1;  if(y>0)      j -= xdim;
	if(Ms_inverse[j]!=0.0) sum_xy = (spin[j] - base);
	j = i;  if(x<xdim-1) j += 1;  if(y>0)      j -= xdim;
	if(Ms_inverse[j]!=0.0) sum_xy += (spin[j] - base);
	j = i;  if(x>0)      j -= 1;  if(y<ydim-1) j += xdim;
	if(Ms_inverse[j]!=0.0) sum_xy += (spin[j] - base);
	j = i;  if(x<xdim-1) j += 1;  if(y<ydim-1) j += xdim;
	if(Ms_inverse[j]!=0.0) sum_xy += (spin[j] - base);
	sum_xy *= wgt_xy;

	// xz-neighbors.  At most 4.
	ThreeVector sum_xz;
	j = i;  if(x>0)      j -= 1;  if(z>0)      j -= xydim;
	if(Ms_inverse[j]!=0.0) sum_xz = (spin[j] - base);
	j = i;  if(x<xdim-1) j += 1;  if(z>0)      j -= xydim;
	if(Ms_inverse[j]!=0.0) sum_xz += (spin[j] - base);
	j = i;  if(x>0)      j -= 1;  if(z<zdim-1) j += xydim;
	if(Ms_inverse[j]!=0.0) sum_xz += (spin[j] - base);
	j = i;  if(x<xdim-1) j += 1;  if(z<zdim-1) j += xydim;
	if(Ms_inverse[j]!=0.0) sum_xz += (spin[j] - base);
	sum_xz *= wgt_xz;

	// yz-neighbors.  At most 4.
	ThreeVector sum_yz;
	j = i;  if(y>0)      j -= xdim;  if(z>0)      j -= xydim;
	if(Ms_inverse[j]!=0.0) sum_yz = (spin[j] - base);
	j = i;  if(y<ydim-1) j += xdim;  if(z>0)      j -= xydim;
	if(Ms_inverse[j]!=0.0) sum_yz += (spin[j] - base);
	j = i;  if(y>0)      j -= xdim;  if(z<zdim-1) j += xydim;
	if(Ms_inverse[j]!=0.0) sum_yz += (spin[j] - base);
	j = i;  if(y<ydim-1) j += xdim;  if(z<zdim-1) j += xydim;
	if(Ms_inverse[j]!=0.0) sum_yz += (spin[j] - base);
	sum_yz *= wgt_yz;

	// Corner neighbors.  At most 8.
	ThreeVector sum_corner;
	UINT4m iz=i;
	if(z>0) iz -= xydim;
	j = iz;  if(x>0)      j -= 1;  if(y>0)      j -= xdim;
	if(Ms_inverse[j]!=0.0) sum_corner = (spin[j] - base);
	j = iz;  if(x<xdim-1) j += 1;  if(y>0)      j -= xdim;
	if(Ms_inverse[j]!=0.0) sum_corner += (spin[j] - base);
	j = iz;  if(x>0)      j -= 1;  if(y<ydim-1) j += xdim;
	if(Ms_inverse[j]!=0.0) sum_corner += (spin[j] - base);
	j = iz;  if(x<xdim-1) j += 1;  if(y<ydim-1) j += xdim;
	if(Ms_inverse[j]!=0.0) sum_corner += (spin[j] - base);
	iz=i;
	if(z<zdim-1) iz += xydim;
	j = iz;  if(x>0)      j -= 1;  if(y>0)      j -= xdim;
	if(Ms_inverse[j]!=0.0) sum_corner += (spin[j] - base);
	j = iz;  if(x<xdim-1) j += 1;  if(y>0)      j -= xdim;
	if(Ms_inverse[j]!=0.0) sum_corner += (spin[j] - base);
	j = iz;  if(x>0)      j -= 1;  if(y<ydim-1) j += xdim;
	if(Ms_inverse[j]!=0.0) sum_corner += (spin[j] - base);
	j = iz;  if(x<xdim-1) j += 1;  if(y<ydim-1) j += xdim;
	if(Ms_inverse[j]!=0.0) sum_corner += (spin[j] - base);
	sum_corner *= wgt_corner;

	// Combine terms
	ThreeVector sum
	  = sum_x  + sum_y  + sum_z
	  + sum_xy + sum_xz + sum_yz
	  + sum_corner;

	field[i] = (hcoef*Msii) * sum;
	energy[i] = (sum * base);
      }
    }
  }
}

void Oxs_UniformExchange::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    throw Oxs_Ext::Error(this,"Import mesh to"
                         " Oxs_UniformExchange::GetEnergy()"
                         " is not an Oxs_RectangularMesh object.");
  }

  // Note: Might want to consider subclassing exchange energies,
  //  to replace this if-block with virtual function pointers.
  if(kernel == NGBR_6) {
    CalcEnergy6Ngbr(spin,Ms_inverse,mesh,energy,field);
  } else if(kernel == NGBR_12) {
    CalcEnergy12Ngbr(spin,Ms_inverse,mesh,energy,field);
  } else if(kernel == NGBR_12_MIRROR) {
    CalcEnergy12NgbrMirror(spin,Ms_inverse,mesh,energy,field);
  } else if(kernel == NGBR_26) {
    CalcEnergy26Ngbr(spin,Ms_inverse,mesh,energy,field);
  } else {
    throw Oxs_Ext::Error(this,"Invalid kernel type detected."
                         "  (Programming error)");
  }

}
