/* FILE: imports.cc              -*-Mode: c++-*-
 *
 *	Replacement implementations for missing Tcl/Tk APIs
 * in old Tcl/Tk releases.
 *
 * Last modified on: $Date: 2002/10/25 20:49:02 $
 * Last modified by: $Author: donahue $
 */

/* Standard libraries */
#include <string.h>

/* Other classes and functions in this extension */
#include "autobuf.h"
#include "imports.h"
#include "nullchannel.h"
/* End includes */     

/* Domain checked atan2 function */
double Oc_Atan2(double y,double x)
{
  // The following #define trick allows atan2 to be defined
  // to be Oc_Atan2 in ocport.h, but allows us to call the system
  // atan2 function from inside this routine.  When the atan2 call
  // is processed by the compiler, if it has previously been
  // #define'd to be Oc_Atan2, then the macro #define'd here kicks
  // in and sets it back to atan2.  Since atan2 is the name of the
  // original macro that is being expanded, macro expansion stops
  // (as per Section 16.3 (?) of the ARM).  If atan2 was not
  // #define'd, or if it was #define'd to something other than
  // Oc_Atan2, then the macro here has no effect.
#define Oc_Atan2(y,x) atan2((y),(x))
  if(x==0.0 && y==0.0) return 0.0;
  return atan2(y,x);
#undef Oc_Atan2
}

// Utility function for computing polynomials of order N
double Oc_Polynomial(double x, const double *coef, unsigned int N) {
    double ans = *coef++;
    while (N--) {
	ans = ans * x + *coef++;
    }
    return ans;
}

// Fallback implementation for those systems where the math
// library fails to provide an erf() function.  On those
// systems a macro in ocport.h will #define erf(x) to be Oc_Erf((x)).
//
// This implementation is one provided by Abdullah Daoud which has
// been passed down in the Institute for Magnetics Research at GWU.
// We should replace it with an implementation from GAMS with known
// numeric properties when we have time.
double Oc_Erf(double x)
{
    const double A[] = {
	 1.061405429,
	-1.453152027,
	 1.421413714,
	-0.284496736,
	 0.254829592,
	 0.0
    };
    const double p	= 0.3275911;
    double t		= 1.0 / (1.0 + p*fabs(x));
    double erfx = 1.0 - exp(-x*x) * Oc_Polynomial(t, A, 5);
    if (x>0) {
        return erfx;
    } else {
        return -erfx;
    }
}

#if (TCL_MAJOR_VERSION < 8)
int
Tcl_GetChannelHandle(Tcl_Channel channel, int dir, ClientData *handlePtr)
{
    Tcl_File tfile=Tcl_GetChannelFile(channel,dir);
    if(tfile==NULL) return TCL_ERROR;
    *handlePtr=Tcl_GetFileInfo(tfile,NULL);
    return TCL_OK;
}
#endif

#if ((TCL_MAJOR_VERSION < 8) || ((TCL_MAJOR_VERSION == 8) \
    && (TCL_MINOR_VERSION == 0) && (TCL_RELEASE_SERIAL < 6)))
char *
Tcl_PkgPresent(Tcl_Interp* interp, char* name, char* version, int exact)
{
    // Need Oc_AutoBuf to support Tcl 7.
    // Starting with Tcl 8.0p0, the interface is CONST
    Oc_AutoBuf ab;
    Tcl_DString buf;
    Tcl_DString loadedVersion;

    /*
     * Use [package provide] to query if the named package is
     * already in the interpreter.
     */ 

    Tcl_DStringInit(&buf);
    Tcl_DStringAppend(&buf, ab("package provide"), -1);
    Tcl_Eval(interp, Tcl_DStringAppendElement(&buf, name));
    Tcl_DStringFree(&buf);

    Tcl_DStringInit(&loadedVersion);
    Tcl_DStringGetResult(interp, &loadedVersion);
    if ( !(Tcl_DStringLength(&loadedVersion)) ) {
        Tcl_DStringFree(&loadedVersion);
        if (version != NULL) {
            Tcl_AppendResult(interp, "package ", name, " ", version,
                    " is not present", (char *) NULL);
        } else {
            Tcl_AppendResult(interp, "package ", name, 
                    " is not present", (char *) NULL);
        }
        return NULL;
    }
    Tcl_DStringFree(&loadedVersion);

    /*
     * At this point we know some version is loaded, so 
     * Tcl_PkgRequire will have no side effect.  Call it
     * to do the rest of the work.
     */

    return Tcl_PkgRequire(interp, name, version, exact);
}
#endif

#if ((TCL_MAJOR_VERSION < 8) \
	|| ((TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION == 0)))
void
Tcl_SaveResult(Tcl_Interp *interp, Tcl_SavedResult *savedPtr)
{
    char *result = Tcl_GetStringResult(interp);

    if (interp->freeProc == TCL_STATIC) {
	/*
	 * String might go away with another Tcl_Eval() call,
	 * so make a copy.
	 */
	int length = strlen(result);
	savedPtr->result = Tcl_Alloc(length + 1);
	strcpy(savedPtr->result, result);
	savedPtr->freeProc = TCL_DYNAMIC;
    } else {
	savedPtr->result = result;
	savedPtr->freeProc = interp->freeProc;
	/*
	 * We've taken responsibility for savedPtr->result, so make
	 * sure the interp doesn't free it for us.
	 */
	interp->freeProc = TCL_STATIC;
    }
    Tcl_ResetResult(interp);
}

void
Tcl_RestoreResult(Tcl_Interp *interp, Tcl_SavedResult *savedPtr)
{
    Tcl_SetResult(interp, savedPtr->result, savedPtr->freeProc);
}

void
Tcl_DiscardResult(Tcl_SavedResult *savedPtr)
{
    if (savedPtr->freeProc == TCL_DYNAMIC) {
	Tcl_Free(savedPtr->result);
    } else {
	(*savedPtr->freeProc)(savedPtr->result);
    }
}
#endif

#if (TK_MAJOR_VERSION < 8)
/*
 * We have not yet implemented a "real" Tcl 7.x version of the null
 * channels, so we have no console channel replacements that will work.
 * If we ever do, remove this restriction.
 * For now, EOF on stdin will kill shells build with Tcl 7.
 */
void Tk_InitConsoleChannels(Tcl_Interp*) {}
#else
/*
 * In Tcl 8.x, a set of null channels can replace the standard channels
 * to prevent EOF on stdin, or other problems from terminating the shell
 * too soon, until a working version is provided by Tk.  
 */

#  if ((SYSTEM_TYPE == WINDOWS) \
	&& (TK_MAJOR_VERSION == 8) && (TK_MINOR_VERSION < 2)) \
	|| (SYSTEM_TYPE == UNIX)
/*
 * Tk_InitConsoleChannels was first made public in Tk 8.2.
 * But on Unix, it breaks [exec].  
 *
 * When it's not provided, or broken, provide the following replacement.
 */

void
Tk_InitConsoleChannels(Tcl_Interp* interp)
{
    Tcl_Channel channel;
    Oc_AutoBuf ab1, ab2;

    channel = Nullchannel_Open();
    if (channel != NULL) {
        Tcl_SetChannelOption(NULL, channel, ab1("-translation"), ab2("lf"));
        Tcl_SetChannelOption(NULL, channel, ab1("-buffering"), ab2("none"));
#if (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION >= 1)
	Tcl_SetChannelOption(NULL, channel, ab1("-encoding"), ab2("utf-8"));
#endif
    }
    Tcl_SetStdChannel(channel, TCL_STDIN);
    Tcl_RegisterChannel(interp, channel);

    channel = Nullchannel_Open();
    if (channel != NULL) {
        Tcl_SetChannelOption(NULL, channel, ab1("-translation"), ab2("lf"));
        Tcl_SetChannelOption(NULL, channel, ab1("-buffering"), ab2("none"));
#if (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION >= 1)
	Tcl_SetChannelOption(NULL, channel, ab1("-encoding"), ab2("utf-8"));
#endif
    }
    Tcl_SetStdChannel(channel, TCL_STDOUT);
    Tcl_RegisterChannel(interp, channel);

#if (SYSTEM_TYPE == WINDOWS)
    /*
     * A fake stderr is what breaks [exec] on Unix.  Don't do it.
     */
    channel = Nullchannel_Open();
    if (channel != NULL) {
        Tcl_SetChannelOption(NULL, channel, ab1("-translation"), ab2("lf"));
        Tcl_SetChannelOption(NULL, channel, ab1("-buffering"), ab2("none"));
#if (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION >= 1)
	Tcl_SetChannelOption(NULL, channel, ab1("-encoding"), ab2("utf-8"));
#endif
    }
    Tcl_SetStdChannel(channel, TCL_STDERR);
    Tcl_RegisterChannel(interp, channel);
#endif

}

#  endif		
#endif

