#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <limits.h>
#include <math.h>
#include <float.h>
#if defined(__unix) || defined(_AIX)
#include <unistd.h>
#endif /* __unix */

/* End includes */

/****************
#define COMPLEX
#define EXTRALONG
****************/

/*
 * If "EXTRALONG" is defined, then types "long long" and "long double"
 * are included.  In particular, the second (long double), may be the
 * 10 byte IEEE 854 (3.4E-4932 to 3.4E+4932) "double-extended-precision"
 * format (as opposed to the 8-byte IEEE 754 "double-precision" format).
 * Your system might have <ieee754.h> and <ieee854.h> include files with
 * more information.
 *
 * Some "magic" values for the IEEE formats
 *
 *               Value                     Internal representation
 *                                         (MSB order, hexadecimal)
 *   4-byte: 2.387939261E-38               01 02 03 04
 *   8-byte: 8.20788039913184000E-304      01 02 03 04 05 06 07 08
 *  10-byte: 7.9695407516944429900E-4855   01 02 83 04 05 06 07 08 09 0A
 *
 *  Note that the third byte in the 10-byte format is '0x83', NOT '0x03'.
 *  This is because (apparently) replacing 0x83 with 0x03 is not a valid
 *  bit pattern for IEEE 854.
 */

/* Debugging routine */
#ifdef __cplusplus
void DumpBytes(char *buf,int len)
#else
void DumpBytes(buf,len)
char *buf;
int len;
#endif
{
  int i;
  for(i=0;i<len;i++) printf("%02X ",(unsigned char)buf[i]);
  printf("\n");
}

#define FillByteOrder(x,i,l) \
   for(i=2,x=1;i<=l;i++) { x<<=8; x|=(unsigned char)i; }


/* IEEE 754 floating point format "magic" strings (Hex 0x1234, 0x12345678) */
#define FLOATCONST(l) (l==4 ? (float)2.387939261E-38 : \
		       (l==8 ? (float)8.20788039913184000E-304 : \
			(float)0.))
#define DOUBLECONST(l) (l==4 ? (double)2.387939261E-38 : \
		       (l==8 ? (double)8.20788039913184000E-304 : \
			(double)0.))
#ifdef EXTRALONG
#define LONGDOUBLECONST(l) (l==4 ? (long double)2.387939261E-38 : \
			 (l==8 ? (long double)8.20788039913184000E-304 : \
			 (l>8 ? (long double)7.9695407516944429900E-4855L : \
			 (long double)0.)))
#endif /* EXTRALONG */


#ifdef __cplusplus
void PrintByteOrder(char *buf,int len)
#else
void PrintByteOrder(buf,len)
char* buf;
int len;
#endif
{
  int i,ch;
  printf("Byte order: ");
  for(i=0;i<len;i++) {
    ch=0x0F & ((unsigned char)buf[i]);
    if(ch==0) printf("x");
    else      printf("%01X",ch);
  }
}

double DblMachEps()
{
  double epsok,eps;
  epsok=2.0;
  eps=epsok/2.0;
  while((1.0+eps)!=1.0 && (1.0-eps)!=1.0) {
    epsok=eps;
    eps/=2.0;
  }
  return epsok;
}

#ifdef EXTRALONG
long double LongDblMachEps()
{
  long double epsok,eps;
  epsok=2.0;
  eps=epsok/2.0;
  while((1.0+eps)!=1.0 && (1.0-eps)!=1.0) {
    epsok=eps;
    eps/=2.0;
  }
  return epsok;
}
#endif /* EXTRALONG */

void Usage()
{
  fprintf(stderr,"Usage: varinfo [-h|--help]"
	  " [--skip-atan2] [--skip-underflow]\n");
  exit(1);
}

#ifdef __cplusplus
int main(int argc,char** argv)
#else
int main(argc,argv)
int argc;
char **argv;
#endif
{
  short s;
  int i;
  long l;
  float f;
  double d;
#ifdef EXTRALONG
  long long ll; long double ld;
#endif /* EXTRALONG */  
#ifdef COMPLEX
  __complex__ int ci;
  __complex__ double cd;
#endif /* COMPLEX */
  size_t st,loop;
  int skip_atan2,skip_underflow;

  /* Check command line flags */
  if(argc>3) Usage();
  skip_atan2=0;
  skip_underflow=0;
  if(argc>1) {
    int i;
    for(i=1;i<argc;i++) {
      if(strcmp("--skip-atan2",argv[i])==0) skip_atan2=1;
      else if(strcmp("--skip-underflow",argv[i])==0) skip_underflow=1;
      else Usage();
    }
  }

  /* Work length info */
  printf("Type        char is %2d bytes wide   ",(int)sizeof(char));
  st=sizeof(char);
  if(st!=1) printf("ERROR: char should be 1 byte wide!\n");
  else      printf("Byte order: 1\n");

  printf("\n");
  printf("Type       short is %2d bytes wide   ",(int)sizeof(short));
  FillByteOrder(s,loop,sizeof(short));
  PrintByteOrder((char *)&s,(int)sizeof(short));      printf("\n");

  printf("Type         int is %2d bytes wide   ",(int)sizeof(int));
  FillByteOrder(i,loop,sizeof(int));
  PrintByteOrder((char *)&i,(int)sizeof(int));        printf("\n");

  printf("Type        long is %2d bytes wide   ",(int)sizeof(long));
  FillByteOrder(l,loop,sizeof(long));
  PrintByteOrder((char *)&l,(int)sizeof(long));       printf("\n");

#ifdef EXTRALONG
  printf("Type   long long is %2d bytes wide   ",sizeof(long long));
  FillByteOrder(ll,loop,sizeof(long long));
  PrintByteOrder((char *)&ll,(int)sizeof(long long)); printf("\n");
#endif /* EXTRALONG */

  printf("\n");
  printf("Type       float is %2d bytes wide   ",(int)sizeof(float));
  f=FLOATCONST(sizeof(float));
  PrintByteOrder((char *)&f,(int)sizeof(float));      printf("\n");

  printf("Type      double is %2d bytes wide   ",(int)sizeof(double));
  d=DOUBLECONST(sizeof(double));
  PrintByteOrder((char *)&d,(int)sizeof(double));     printf("\n");

#ifdef EXTRALONG
  printf("Type long double is %2d bytes wide   ",sizeof(long double));
  ld=LONGDOUBLECONST(sizeof(long double));
  PrintByteOrder((char *)&ld,(int)sizeof(long double));  printf("\n");
#endif /* EXTRALONG */

  printf("\n");
  printf("Type      void * is %2d bytes wide\n",(int)sizeof(void *));

#ifdef COMPLEX
  printf("\n");
  printf("Type __complex__    int is %2d bytes wide\n",
	 sizeof(__complex__ int));
  printf("Type __complex__ double is %2d bytes wide\n",
	 sizeof(__complex__ double));
#endif /* COMPLEX */

  /* Byte order info */
  printf("\n");  fflush(stdout);
  st=sizeof(double);
  if(st!=8) {
    fprintf(stderr,"Can't test byte order; sizeof(double)!=8\n");
  }
  else {
    union {
      double x;
      unsigned char c[8];
    } bytetest;
    double xpattern=1.234;
    unsigned char lsbpattern[9],msbpattern[9];
    strncpy((char *)lsbpattern,"\130\071\264\310\166\276\363\077",
	    sizeof(lsbpattern));
    strncpy((char *)msbpattern,"\077\363\276\166\310\264\071\130",
	    sizeof(msbpattern));

    /* The bit pattern for 1.234 on a little endian machine (e.g.,
     * Intel's x86 series and Digital's AXP) should be
     * 58 39 B4 C8 76 BE F3 3F, while on a big endian machine it is
     * 3F F3 BE 76 C8 B4 39 58.  Examples of big endian machines
     * include ones using the MIPS R4000 and R8000 series chips,
     * for example SGI's.  Note: At least some versions of the Sun
     * 'cc' compiler apparently don't grok '\xXX' hex notation.  The
     * above octal constants are equivalent to the aforementioned
     * hex strings.
     */
    bytetest.x=xpattern;

#ifdef DEBUG    
    printf("sizeof(lsbpattern)=%d\n",sizeof(lsbpattern));
    printf("lsb pattern="); DumpBytes(lsbpattern,8);
    printf("msb pattern="); DumpBytes(msbpattern,8);
    printf("  x pattern="); DumpBytes(bytetest.c,8);
#endif /* DEBUG */

    /* Floating point format */
    if(strncmp((char *)bytetest.c,(char *)lsbpattern,8)==0) {
      printf("Floating point format is IEEE in LSB (little endian) order\n");
    }
    else if(strncmp((char *)bytetest.c,(char *)msbpattern,8)==0) {
      printf("Floating point format is IEEE in MSB (big endian) order\n");
    }
    else {
      printf("Floating point format is either unknown or uses an irregular byte order\n");
    }
  }

  /* Additional system-specific information (ANSI) */
  printf("\nWidth of clock_t variable: %7lu bytes\n",
	 (unsigned long)sizeof(clock_t));
#ifdef CLOCKS_PER_SEC
  printf(  "           CLOCKS_PER_SEC: %7lu\n",
	   (unsigned long)CLOCKS_PER_SEC);
#else
  printf(  "           CLOCKS_PER_SEC: <not defined>\n");
#endif
#ifdef CLK_TCK
  printf(  "                  CLK_TCK: %7lu\n",(unsigned long)CLK_TCK);
#else
  printf(  "                  CLK_TCK: <not defined>\n");
#endif

  printf("\nFLT_EPSILON: %.9g\n",FLT_EPSILON);
  printf("SQRT_FLT_EPSILON: %.9g\n",sqrt(FLT_EPSILON));
  printf("DBL_EPSILON: %.17g\n",DBL_EPSILON);
  printf("SQRT_DBL_EPSILON: %.17g\n",sqrt(DBL_EPSILON));
  printf("CUBE_ROOT_DBL_EPSILON: %.17g\n",pow(DBL_EPSILON,1./3.));

  printf("\nCalculated Double Epsilon: %.17g\n",DblMachEps());
#ifdef EXTRALONG
  printf("Calculated Long Double Epsilon: %.17lg\n",LongDblMachEps());
#endif /* EXTRALONG */

  printf("\nDBL_MIN: %.17g\n",DBL_MIN);
  printf("DBL_MAX: %.17g\n",DBL_MAX);

  if(!skip_atan2) {
    printf("\nReturn value from atan2(0,0): %g\n",atan2(0.,0.));
  }

  if(!skip_underflow) {
    d = -999999.;
    errno = 0;
    d = exp(d);
    if(d!=0.0) {
      fprintf(stderr,"\nERROR: UNDERFLOW TEST FAILURE\n\n");
      exit(1);
    } 
    if(errno==0) printf("\nErrno not set on underflow\n");
    else         printf("\nErrno set on underflow to %d\n",errno);
  }

  return 0;
}
